# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Hooks for Data Lineage commands."""

from apitools.base.py import encoding
from apitools.base.py import extra_types
from googlecloudsdk.core import yaml


def SetProcessNameInBody(ref, args, req):
  """Set the process name in the request body."""
  del args
  if req.googleCloudDatacatalogLineageV1Process is not None:
    req.googleCloudDatacatalogLineageV1Process.name = ref.RelativeName()
  return req


def ParseAttributeValue(value):
  """Parse attribute value as YAML/JSON and convert to JsonValue message."""
  try:
    parsed_python_value = yaml.load(value)
  except Exception:  # pylint: disable=broad-except
    parsed_python_value = value
  return encoding.PyValueToMessage(extra_types.JsonValue, parsed_python_value)



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Waiter utility for api_lib.util.waiter.py."""


import logging
from apitools.base.py import exceptions as apitools_exceptions
from googlecloudsdk.api_lib.dataproc import exceptions
from googlecloudsdk.api_lib.dataproc import util
from googlecloudsdk.api_lib.dataproc.poller import cloud_console_url_helper
from googlecloudsdk.api_lib.util import waiter
from googlecloudsdk.core import log


class RmBatchPoller(waiter.OperationPoller):
  """Poller for resource manager batches.

  This should be used for spark version 3+, and Ray version 1+.
  """

  def __init__(self, dataproc):
    self.dataproc = dataproc
    self.fist_tick_message_printed = False

  def IsDone(self, batch):
    """See base class."""
    if batch and batch.state in (
        self.dataproc.messages.Batch.StateValueValuesEnum.SUCCEEDED,
        self.dataproc.messages.Batch.StateValueValuesEnum.CANCELLED,
        self.dataproc.messages.Batch.StateValueValuesEnum.FAILED,
    ):
      return True
    return False

  def Poll(self, batch_ref):
    """See base class."""
    request = self.dataproc.messages.DataprocProjectsLocationsBatchesGetRequest(
        name=batch_ref
    )
    try:
      return self.dataproc.client.projects_locations_batches.Get(request)
    except apitools_exceptions.HttpError as error:
      log.warning('Get Batch failed:\n{}'.format(error))
      if util.IsClientHttpException(error):
        # Stop polling if encounter client Http error (4xx).
        raise

  def GetResult(self, batch):
    """Handles errors.

    Error handling for batch jobs. This happen after the batch reaches one of
    the complete states.

    Overrides.

    Args:
      batch: The batch resource.

    Returns:
      None. The result is directly output to log.err.

    Raises:
      JobTimeoutError: When waiter timed out.
      JobError: When remote batch job is failed.
    """
    if not batch:
      # Batch resource is None but polling is considered done.
      # This only happens when the waiter timed out.
      raise exceptions.JobTimeoutError('Timed out while waiting for batch job.')

    if (
        batch.state
        == self.dataproc.messages.Batch.StateValueValuesEnum.CANCELLED
    ):
      log.warning('Batch job is CANCELLED.')
    elif (
        batch.state == self.dataproc.messages.Batch.StateValueValuesEnum.FAILED
    ):
      err_message = 'Batch job is FAILED.'
      if batch.stateMessage:
        err_message = '{} Detail: {}'.format(err_message, batch.stateMessage)
        if err_message[-1] != '.':
          err_message += '.'
      raise exceptions.JobError(err_message)

    # Nothing to return, since the result is directly output to users.
    return None

  def TrackerUpdateFunction(self, tracker, poll_result, status):
    """Prints links to cloud console after the first success pull."""
    if not self.fist_tick_message_printed:
      self.fist_tick_message_printed = True
      cloud_logging_url = cloud_console_url_helper.get_batch_logging_url(
          poll_result
      )
      dataproc_batch_url = cloud_console_url_helper.get_dataproc_batch_url(
          poll_result
      )
      log_level = log.GetVerbosity()
      log.SetVerbosity(logging.INFO)
      log.info(
          'Please check the driver output in Cloud Logging: %s. (The log can'
          ' take a few minutes to show up.) You can visit the batch resource'
          ' at %s',
          cloud_logging_url,
          dataproc_batch_url,
      )
      log.info('Waiting for the batch to complete.')
      log.SetVerbosity(log_level)



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Database Migration Service conversion workspaces Base Client."""

import abc
from typing import Iterable, Optional, TYPE_CHECKING

from googlecloudsdk.api_lib.database_migration import api_util
from googlecloudsdk.calliope import base
from googlecloudsdk.generated_clients.apis.datamigration.v1 import datamigration_v1_client as client

# pylint: disable=g-bad-import-order
if TYPE_CHECKING:
  from googlecloudsdk.api_lib.database_migration.conversion_workspaces import conversion_workspaces_client  # pylint: disable=g-import-not-at-top


class BaseConversionWorkspacesClient(abc.ABC):
  """Base Client for Conversion Workspaces APIs.

  This class is the base class for the conversion workspaces clients and
  provides the common services used by the clients in order to send API
  requests.

  Each client inheriting from this class handles a specific part of the API, and
  might need to call other clients in order to complete its
  functionality. Accessing other clients is done through the parent_client
  property.

  Attributes:
    release_track: The release track of the client, controlling the API version
      to use.
    parent_client: The parent client of the conversion workspaces client.
    location: The location of the resources.
    client: The client used to send API requests.
    messages: The messages module used to construct API requests.
  """

  def __init__(
      self,
      release_track: base.ReleaseTrack,
      parent_client: 'conversion_workspaces_client.ConversionWorkspacesClient',
      location: Optional[str] = None,
  ):
    """Initializes the instance with an API client based on the release track.

    Args:
      release_track: The release track of the client, controlling the API
        version to use.
      parent_client: The parent client of the conversion workspaces client.
      location: The location of the resources.
    """

    self.release_track = release_track
    self.parent_client = parent_client

    self.client: client.DatamigrationV1 = api_util.GetClientInstance(
        release_track=release_track,
        location=location,
    )
    self.messages = api_util.GetMessagesModule(release_track=release_track)

  @property
  def cw_service(
      self,
  ) -> client.DatamigrationV1.ProjectsLocationsConversionWorkspacesService:
    """Returns the conversion workspaces service."""
    return self.client.projects_locations_conversionWorkspaces

  @property
  def mapping_rules_service(
      self,
  ) -> (
      client.DatamigrationV1.ProjectsLocationsConversionWorkspacesMappingRulesService
  ):
    """Returns the mapping rules service."""
    return self.client.projects_locations_conversionWorkspaces_mappingRules

  @property
  def location_service(
      self,
  ) -> client.DatamigrationV1.ProjectsLocationsService:
    """Returns the location service."""
    return self.client.projects_locations

  def CombineFilters(
      self,
      *filter_exprs: Iterable[Optional[str]],
  ) -> Optional[str]:
    """Combine filter expression with global filter.

    Args:
      *filter_exprs: Filter expressions to combine.

    Returns:
      Combined filter expression (or None if no filter expressions are
      provided).
    """

    filter_exprs = tuple(
        filter(
            lambda filter_expr: filter_expr and filter_expr != '*',
            filter_exprs,
        )
    )

    if not filter_exprs:
      return None
    if len(filter_exprs) == 1:
      return filter_exprs[0]

    return ' AND '.join(
        map(lambda filter_expr: f'({filter_expr})', filter_exprs)
    )



"""Generated message classes for edgecontainer version v1beta.

"""
# NOTE: This file is autogenerated and should not be edited by hand.

from apitools.base.protorpclite import messages as _messages
from apitools.base.py import encoding
from apitools.base.py import extra_types


package = 'edgecontainer'


class Authorization(_messages.Message):
  r"""RBAC policy that will be applied and managed by GEC.

  Fields:
    adminUsers: Required. User that will be granted the cluster-admin role on
      the cluster, providing full access to the cluster. Currently, this is a
      singular field, but will be expanded to allow multiple admins in the
      future.
  """

  adminUsers = _messages.MessageField('ClusterUser', 1)


class CancelOperationRequest(_messages.Message):
  r"""The request message for Operations.CancelOperation."""


class ChannelConfig(_messages.Message):
  r"""Configuration for a release channel.

  Fields:
    defaultVersion: Output only. Default version for this release channel,
      e.g.: "1.4.0".
  """

  defaultVersion = _messages.StringField(1)


class CloudRouter(_messages.Message):
  r"""The Cloud Router info.

  Fields:
    name: The associated Cloud Router name.
  """

  name = _messages.StringField(1)


class CloudVpn(_messages.Message):
  r"""The Cloud VPN info.

  Fields:
    gateway: The created Cloud VPN gateway name.
  """

  gateway = _messages.StringField(1)


class Cluster(_messages.Message):
  r"""A Google Distributed Cloud Edge Kubernetes cluster.

  Enums:
    ClusterTypeValueValuesEnum: Optional. Cluster Type to specify if the
      cluster is BAREMETAL or VIRTUAL
    ReleaseChannelValueValuesEnum: Optional. The release channel a cluster is
      subscribed to.
    StatusValueValuesEnum: Output only. The current status of the cluster.

  Messages:
    LabelsValue: Labels associated with this resource.

  Fields:
    authorization: Required. Immutable. RBAC policy that will be applied and
      managed by GEC.
    clusterCaCertificate: Output only. The PEM-encoded public certificate of
      the cluster's CA.
    clusterType: Optional. Cluster Type to specify if the cluster is BAREMETAL
      or VIRTUAL
    connectionState: Output only. The current connection state of the cluster.
    containerRuntimeConfig: Optional. The container runtime config of the
      cluster.
    controlPlane: Optional. The configuration of the cluster control plane.
    controlPlaneEncryption: Optional. Remote control plane disk encryption
      options. This field is only used when enabling CMEK support.
    controlPlaneVersion: Output only. The control plane release version
    createTime: Output only. The time when the cluster was created.
    defaultMaxPodsPerNode: Optional. The default maximum number of pods per
      node used if a maximum value is not specified explicitly for a node pool
      in this cluster. If unspecified, the Kubernetes default value will be
      used.
    enableClusterIsolation: Optional. This denotes if the cluster is required
      to be isolated. go/cluster-isolation-in-gdcc-cluster
    enableRemoteBackup: Optional. If true, the remote backup/restore feature
      will be enabled for this cluster.
    endpoint: Output only. The IP address of the Kubernetes API server.
    externalLoadBalancerAddressPools: Optional. External load balancer pools
      for cluster.
    externalLoadBalancerIpv4AddressPools: Optional. IPv4 address pools for
      cluster data plane external load balancing.
    externalLoadBalancerIpv6AddressPools: Optional. IPv6 address pools for
      cluster data plane external load balancing.
    fleet: Required. Fleet configuration.
    googleGroupAuthentication: Optional. The Google Group authentication
      config of the cluster.
    labels: Labels associated with this resource.
    maintenanceEvents: Output only. All the maintenance events scheduled for
      the cluster, including the ones ongoing, planned for the future and done
      in the past (up to 90 days).
    maintenancePolicy: Optional. Cluster-wide maintenance policy
      configuration.
    name: Required. The resource name of the cluster.
    networking: Required. Cluster-wide networking configuration.
    nodeVersion: Output only. The lowest release version among all worker
      nodes. This field can be empty if the cluster does not have any worker
      nodes.
    port: Output only. The port number of the Kubernetes API server.
    releaseChannel: Optional. The release channel a cluster is subscribed to.
    status: Output only. The current status of the cluster.
    survivabilityConfig: Optional. Configuration of the cluster survivability,
      e.g., for the case when network connectivity is lost. Note: This only
      applies to local control plane clusters.
    systemAddonsConfig: Optional. The configuration of the system add-ons.
    targetVersion: Optional. The target cluster version. For example: "1.5.0".
    updateTime: Output only. The time when the cluster was last updated.
    upgradeSettings: Optional. Upgrade settings for the cluster.
    zoneStorageEncryption: Optional. The zone storage encryption configuration
  """

  class ClusterTypeValueValuesEnum(_messages.Enum):
    r"""Optional. Cluster Type to specify if the cluster is BAREMETAL or
    VIRTUAL

    Values:
      CLUSTER_TYPE_UNSPECIFIED: Unspecified cluster type
      BAREMETAL: Cluster is a baremetal cluster
      VIRTUAL: Cluster is a virtual cluster
    """
    CLUSTER_TYPE_UNSPECIFIED = 0
    BAREMETAL = 1
    VIRTUAL = 2

  class ReleaseChannelValueValuesEnum(_messages.Enum):
    r"""Optional. The release channel a cluster is subscribed to.

    Values:
      RELEASE_CHANNEL_UNSPECIFIED: Unspecified release channel. This will
        default to the REGULAR channel.
      NONE: No release channel.
      REGULAR: Regular release channel.
    """
    RELEASE_CHANNEL_UNSPECIFIED = 0
    NONE = 1
    REGULAR = 2

  class StatusValueValuesEnum(_messages.Enum):
    r"""Output only. The current status of the cluster.

    Values:
      STATUS_UNSPECIFIED: Status unknown.
      PROVISIONING: The cluster is being created.
      RUNNING: The cluster is created and fully usable.
      DELETING: The cluster is being deleted.
      ERROR: The status indicates that some errors occurred while
        reconciling/deleting the cluster.
      RECONCILING: The cluster is undergoing some work such as version
        upgrades, etc.
    """
    STATUS_UNSPECIFIED = 0
    PROVISIONING = 1
    RUNNING = 2
    DELETING = 3
    ERROR = 4
    RECONCILING = 5

  @encoding.MapUnrecognizedFields('additionalProperties')
  class LabelsValue(_messages.Message):
    r"""Labels associated with this resource.

    Messages:
      AdditionalProperty: An additional property for a LabelsValue object.

    Fields:
      additionalProperties: Additional properties of type LabelsValue
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a LabelsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  authorization = _messages.MessageField('Authorization', 1)
  clusterCaCertificate = _messages.StringField(2)
  clusterType = _messages.EnumField('ClusterTypeValueValuesEnum', 3)
  connectionState = _messages.MessageField('ConnectionState', 4)
  containerRuntimeConfig = _messages.MessageField('ContainerRuntimeConfig', 5)
  controlPlane = _messages.MessageField('ControlPlane', 6)
  controlPlaneEncryption = _messages.MessageField('ControlPlaneEncryption', 7)
  controlPlaneVersion = _messages.StringField(8)
  createTime = _messages.StringField(9)
  defaultMaxPodsPerNode = _messages.IntegerField(10, variant=_messages.Variant.INT32)
  enableClusterIsolation = _messages.BooleanField(11)
  enableRemoteBackup = _messages.BooleanField(12)
  endpoint = _messages.StringField(13)
  externalLoadBalancerAddressPools = _messages.MessageField('ExternalLoadBalancerPool', 14, repeated=True)
  externalLoadBalancerIpv4AddressPools = _messages.StringField(15, repeated=True)
  externalLoadBalancerIpv6AddressPools = _messages.StringField(16, repeated=True)
  fleet = _messages.MessageField('Fleet', 17)
  googleGroupAuthentication = _messages.MessageField('GoogleGroupAuthenticationConfig', 18)
  labels = _messages.MessageField('LabelsValue', 19)
  maintenanceEvents = _messages.MessageField('MaintenanceEvent', 20, repeated=True)
  maintenancePolicy = _messages.MessageField('MaintenancePolicy', 21)
  name = _messages.StringField(22)
  networking = _messages.MessageField('ClusterNetworking', 23)
  nodeVersion = _messages.StringField(24)
  port = _messages.IntegerField(25, variant=_messages.Variant.INT32)
  releaseChannel = _messages.EnumField('ReleaseChannelValueValuesEnum', 26)
  status = _messages.EnumField('StatusValueValuesEnum', 27)
  survivabilityConfig = _messages.MessageField('SurvivabilityConfig', 28)
  systemAddonsConfig = _messages.MessageField('SystemAddonsConfig', 29)
  targetVersion = _messages.StringField(30)
  updateTime = _messages.StringField(31)
  upgradeSettings = _messages.MessageField('UpgradeSettings', 32)
  zoneStorageEncryption = _messages.MessageField('ZoneStorageEncryption', 33)


class ClusterNetworking(_messages.Message):
  r"""Cluster-wide networking configuration.

  Enums:
    NetworkTypeValueValuesEnum: Output only. IP addressing type of this
      cluster i.e. SINGLESTACK_V4 vs DUALSTACK_V4_V6

  Fields:
    clusterIpv4CidrBlocks: Required. All pods in the cluster are assigned an
      RFC1918 IPv4 address from these blocks. Only a single block is
      supported. This field cannot be changed after creation.
    clusterIpv6CidrBlocks: Immutable. If specified, dual stack mode is enabled
      and all pods in the cluster are assigned an IPv6 address from these
      blocks alongside from an IPv4 address. Only a single block is supported.
      This field cannot be changed after creation.
    networkType: Output only. IP addressing type of this cluster i.e.
      SINGLESTACK_V4 vs DUALSTACK_V4_V6
    servicesIpv4CidrBlocks: Required. All services in the cluster are assigned
      an RFC1918 IPv4 address from these blocks. Only a single block is
      supported. This field cannot be changed after creation.
    servicesIpv6CidrBlocks: Immutable. If specified, dual stack mode is
      enabled and all services in the cluster are assigned an IPv6 address
      from these blocks alongside from an IPv4 address. Only a single block is
      supported. This field cannot be changed after creation.
  """

  class NetworkTypeValueValuesEnum(_messages.Enum):
    r"""Output only. IP addressing type of this cluster i.e. SINGLESTACK_V4 vs
    DUALSTACK_V4_V6

    Values:
      NETWORK_TYPE_UNSPECIFIED: Unknown cluster type
      SINGLESTACK_V4: SingleStack v4 address only cluster
      DUALSTACK_V4_V6: DualStack cluster - support v4 and v6 address
    """
    NETWORK_TYPE_UNSPECIFIED = 0
    SINGLESTACK_V4 = 1
    DUALSTACK_V4_V6 = 2

  clusterIpv4CidrBlocks = _messages.StringField(1, repeated=True)
  clusterIpv6CidrBlocks = _messages.StringField(2, repeated=True)
  networkType = _messages.EnumField('NetworkTypeValueValuesEnum', 3)
  servicesIpv4CidrBlocks = _messages.StringField(4, repeated=True)
  servicesIpv6CidrBlocks = _messages.StringField(5, repeated=True)


class ClusterUser(_messages.Message):
  r"""A user principal for an RBAC policy.

  Fields:
    username: Required. An active Google username.
  """

  username = _messages.StringField(1)


class ConfigData(_messages.Message):
  r"""Config data holds all the config related data for the zone.

  Fields:
    availableExternalLbPoolsIpv4: list of available v4 ip pools for external
      loadbalancer
    availableExternalLbPoolsIpv6: list of available v6 ip pools for external
      loadbalancer
  """

  availableExternalLbPoolsIpv4 = _messages.StringField(1, repeated=True)
  availableExternalLbPoolsIpv6 = _messages.StringField(2, repeated=True)


class ConnectionState(_messages.Message):
  r"""ConnectionState holds the current connection state from the cluster to
  Google.

  Enums:
    StateValueValuesEnum: Output only. The current connection state.

  Fields:
    state: Output only. The current connection state.
    updateTime: Output only. The time when the connection state was last
      changed.
  """

  class StateValueValuesEnum(_messages.Enum):
    r"""Output only. The current connection state.

    Values:
      STATE_UNSPECIFIED: Unknown connection state.
      DISCONNECTED: This cluster is currently disconnected from Google.
      CONNECTED: This cluster is currently connected to Google.
      CONNECTED_AND_SYNCING: This cluster is currently connected to Google,
        but may have recently reconnected after a disconnection. It is still
        syncing back.
    """
    STATE_UNSPECIFIED = 0
    DISCONNECTED = 1
    CONNECTED = 2
    CONNECTED_AND_SYNCING = 3

  state = _messages.EnumField('StateValueValuesEnum', 1)
  updateTime = _messages.StringField(2)


class ContainerRuntimeConfig(_messages.Message):
  r"""Container runtime config of the cluster.

  Enums:
    DefaultContainerRuntimeValueValuesEnum: Optional. The default container
      runtime to be configured in the cluster.

  Fields:
    defaultContainerRuntime: Optional. The default container runtime to be
      configured in the cluster.
  """

  class DefaultContainerRuntimeValueValuesEnum(_messages.Enum):
    r"""Optional. The default container runtime to be configured in the
    cluster.

    Values:
      DEFAULT_CONTAINER_RUNTIME_UNSPECIFIED: Container runtime not specified.
      RUNC: Use runc as the default container runtime in the cluster.
      GVISOR: Use gVisor as the default container runtime in the cluster.
    """
    DEFAULT_CONTAINER_RUNTIME_UNSPECIFIED = 0
    RUNC = 1
    GVISOR = 2

  defaultContainerRuntime = _messages.EnumField('DefaultContainerRuntimeValueValuesEnum', 1)


class ControlPlane(_messages.Message):
  r"""Configuration of the cluster control plane.

  Fields:
    local: Local control plane configuration. Warning: Local control plane
      clusters must be created in their own project. Local control plane
      clusters cannot coexist in the same project with any other type of
      clusters, including non-GDCE clusters. Mixing local control plane GDCE
      clusters with any other type of clusters in the same project can result
      in data loss.
    remote: Remote control plane configuration.
  """

  local = _messages.MessageField('Local', 1)
  remote = _messages.MessageField('Remote', 2)


class ControlPlaneEncryption(_messages.Message):
  r"""Configuration for Customer-managed KMS key support for control plane
  nodes.

  Enums:
    KmsKeyStateValueValuesEnum: Output only. Availability of the Cloud KMS
      CryptoKey. If not `KEY_AVAILABLE`, then nodes may go offline as they
      cannot access their local data. This can be caused by a lack of
      permissions to use the key, or if the key is disabled or deleted.
    ResourceStateValueValuesEnum: Output only. The current resource state
      associated with the cmek.

  Fields:
    kmsKey: Optional. The Cloud KMS CryptoKey e.g. projects/{project}/location
      s/{location}/keyRings/{keyRing}/cryptoKeys/{cryptoKey} to use for
      protecting control plane disks. If not specified, a Google-managed key
      will be used instead.
    kmsKeyActiveVersion: Output only. The Cloud KMS CryptoKeyVersion currently
      in use for protecting control plane disks. Only applicable if kms_key is
      set.
    kmsKeyState: Output only. Availability of the Cloud KMS CryptoKey. If not
      `KEY_AVAILABLE`, then nodes may go offline as they cannot access their
      local data. This can be caused by a lack of permissions to use the key,
      or if the key is disabled or deleted.
    kmsStatus: Output only. Error status returned by Cloud KMS when using this
      key. This field may be populated only if `kms_key_state` is not
      `KMS_KEY_STATE_KEY_AVAILABLE`. If populated, this field contains the
      error status reported by Cloud KMS.
    resourceState: Output only. The current resource state associated with the
      cmek.
  """

  class KmsKeyStateValueValuesEnum(_messages.Enum):
    r"""Output only. Availability of the Cloud KMS CryptoKey. If not
    `KEY_AVAILABLE`, then nodes may go offline as they cannot access their
    local data. This can be caused by a lack of permissions to use the key, or
    if the key is disabled or deleted.

    Values:
      KMS_KEY_STATE_UNSPECIFIED: Unspecified.
      KMS_KEY_STATE_KEY_AVAILABLE: The key is available for use, and dependent
        resources should be accessible.
      KMS_KEY_STATE_KEY_UNAVAILABLE: The key is unavailable for an unspecified
        reason. Dependent resources may be inaccessible.
    """
    KMS_KEY_STATE_UNSPECIFIED = 0
    KMS_KEY_STATE_KEY_AVAILABLE = 1
    KMS_KEY_STATE_KEY_UNAVAILABLE = 2

  class ResourceStateValueValuesEnum(_messages.Enum):
    r"""Output only. The current resource state associated with the cmek.

    Values:
      RESOURCE_STATE_UNSPECIFIED: Default value.
      RESOURCE_STATE_LOCK_DOWN: The resource is in LOCK DOWN state.
      RESOURCE_STATE_LOCK_DOWN_PENDING: The resource is pending lock down.
    """
    RESOURCE_STATE_UNSPECIFIED = 0
    RESOURCE_STATE_LOCK_DOWN = 1
    RESOURCE_STATE_LOCK_DOWN_PENDING = 2

  kmsKey = _messages.StringField(1)
  kmsKeyActiveVersion = _messages.StringField(2)
  kmsKeyState = _messages.EnumField('KmsKeyStateValueValuesEnum', 3)
  kmsStatus = _messages.MessageField('Status', 4)
  resourceState = _messages.EnumField('ResourceStateValueValuesEnum', 5)


class Details(_messages.Message):
  r"""The created connection details.

  Enums:
    StateValueValuesEnum: The state of this connection.

  Fields:
    cloudRouter: The Cloud Router info.
    cloudVpns: Each connection has multiple Cloud VPN gateways.
    error: The error message. This is only populated when state=ERROR.
    state: The state of this connection.
  """

  class StateValueValuesEnum(_messages.Enum):
    r"""The state of this connection.

    Values:
      STATE_UNSPECIFIED: Unknown.
      STATE_CONNECTED: Connected.
      STATE_CONNECTING: Still connecting.
      STATE_ERROR: Error occurred.
    """
    STATE_UNSPECIFIED = 0
    STATE_CONNECTED = 1
    STATE_CONNECTING = 2
    STATE_ERROR = 3

  cloudRouter = _messages.MessageField('CloudRouter', 1)
  cloudVpns = _messages.MessageField('CloudVpn', 2, repeated=True)
  error = _messages.StringField(3)
  state = _messages.EnumField('StateValueValuesEnum', 4)


class EdgecontainerProjectsLocationsClustersCreateRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsClustersCreateRequest object.

  Fields:
    cluster: A Cluster resource to be passed as the request body.
    clusterId: Required. A client-specified unique identifier for the cluster.
    parent: Required. The parent location where this cluster will be created.
    requestId: A unique identifier for this request. Restricted to 36 ASCII
      characters. A random UUID is recommended. This request is only
      idempotent if `request_id` is provided.
  """

  cluster = _messages.MessageField('Cluster', 1)
  clusterId = _messages.StringField(2)
  parent = _messages.StringField(3, required=True)
  requestId = _messages.StringField(4)


class EdgecontainerProjectsLocationsClustersDeleteRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsClustersDeleteRequest object.

  Fields:
    name: Required. The resource name of the cluster.
    requestId: A unique identifier for this request. Restricted to 36 ASCII
      characters. A random UUID is recommended. This request is only
      idempotent if `request_id` is provided.
  """

  name = _messages.StringField(1, required=True)
  requestId = _messages.StringField(2)


class EdgecontainerProjectsLocationsClustersGenerateAccessTokenRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsClustersGenerateAccessTokenRequest
  object.

  Fields:
    cluster: Required. The resource name of the cluster.
  """

  cluster = _messages.StringField(1, required=True)


class EdgecontainerProjectsLocationsClustersGenerateOfflineCredentialRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsClustersGenerateOfflineCredentialRequest
  object.

  Fields:
    cluster: Required. The resource name of the cluster.
  """

  cluster = _messages.StringField(1, required=True)


class EdgecontainerProjectsLocationsClustersGetRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsClustersGetRequest object.

  Fields:
    name: Required. The resource name of the cluster.
  """

  name = _messages.StringField(1, required=True)


class EdgecontainerProjectsLocationsClustersListRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsClustersListRequest object.

  Fields:
    filter: Only resources matching this filter will be listed.
    orderBy: Specifies the order in which resources will be listed.
    pageSize: The maximum number of resources to list.
    pageToken: A page token received from previous list request. A page token
      received from previous list request.
    parent: Required. The parent location, which owns this collection of
      clusters.
  """

  filter = _messages.StringField(1)
  orderBy = _messages.StringField(2)
  pageSize = _messages.IntegerField(3, variant=_messages.Variant.INT32)
  pageToken = _messages.StringField(4)
  parent = _messages.StringField(5, required=True)


class EdgecontainerProjectsLocationsClustersNodePoolsCreateRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsClustersNodePoolsCreateRequest object.

  Fields:
    nodePool: A NodePool resource to be passed as the request body.
    nodePoolId: Required. A client-specified unique identifier for the node
      pool.
    parent: Required. The parent cluster where this node pool will be created.
    requestId: A unique identifier for this request. Restricted to 36 ASCII
      characters. A random UUID is recommended. This request is only
      idempotent if `request_id` is provided.
  """

  nodePool = _messages.MessageField('NodePool', 1)
  nodePoolId = _messages.StringField(2)
  parent = _messages.StringField(3, required=True)
  requestId = _messages.StringField(4)


class EdgecontainerProjectsLocationsClustersNodePoolsDeleteRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsClustersNodePoolsDeleteRequest object.

  Fields:
    name: Required. The resource name of the node pool.
    requestId: A unique identifier for this request. Restricted to 36 ASCII
      characters. A random UUID is recommended. This request is only
      idempotent if `request_id` is provided.
  """

  name = _messages.StringField(1, required=True)
  requestId = _messages.StringField(2)


class EdgecontainerProjectsLocationsClustersNodePoolsGetRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsClustersNodePoolsGetRequest object.

  Fields:
    name: Required. The resource name of the node pool.
  """

  name = _messages.StringField(1, required=True)


class EdgecontainerProjectsLocationsClustersNodePoolsListRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsClustersNodePoolsListRequest object.

  Fields:
    filter: Only resources matching this filter will be listed.
    orderBy: Specifies the order in which resources will be listed.
    pageSize: The maximum number of resources to list.
    pageToken: A page token received from previous list request.
    parent: Required. The parent cluster, which owns this collection of node
      pools.
  """

  filter = _messages.StringField(1)
  orderBy = _messages.StringField(2)
  pageSize = _messages.IntegerField(3, variant=_messages.Variant.INT32)
  pageToken = _messages.StringField(4)
  parent = _messages.StringField(5, required=True)


class EdgecontainerProjectsLocationsClustersNodePoolsPatchRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsClustersNodePoolsPatchRequest object.

  Fields:
    name: Required. The resource name of the node pool.
    nodePool: A NodePool resource to be passed as the request body.
    requestId: A unique identifier for this request. Restricted to 36 ASCII
      characters. A random UUID is recommended. This request is only
      idempotent if `request_id` is provided.
    updateMask: Field mask is used to specify the fields to be overwritten in
      the NodePool resource by the update. The fields specified in the
      update_mask are relative to the resource, not the full request. A field
      will be overwritten if it is in the mask. If the user does not provide a
      mask then all fields will be overwritten.
  """

  name = _messages.StringField(1, required=True)
  nodePool = _messages.MessageField('NodePool', 2)
  requestId = _messages.StringField(3)
  updateMask = _messages.StringField(4)


class EdgecontainerProjectsLocationsClustersPatchRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsClustersPatchRequest object.

  Fields:
    cluster: A Cluster resource to be passed as the request body.
    name: Required. The resource name of the cluster.
    requestId: A unique identifier for this request. Restricted to 36 ASCII
      characters. A random UUID is recommended. This request is only
      idempotent if `request_id` is provided.
    updateMask: Field mask is used to specify the fields to be overwritten in
      the Cluster resource by the update. The fields specified in the
      update_mask are relative to the resource, not the full request. A field
      will be overwritten if it is in the mask. If the user does not provide a
      mask then all fields will be overwritten.
  """

  cluster = _messages.MessageField('Cluster', 1)
  name = _messages.StringField(2, required=True)
  requestId = _messages.StringField(3)
  updateMask = _messages.StringField(4)


class EdgecontainerProjectsLocationsClustersUpgradeRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsClustersUpgradeRequest object.

  Fields:
    name: Required. The resource name of the cluster.
    upgradeClusterRequest: A UpgradeClusterRequest resource to be passed as
      the request body.
  """

  name = _messages.StringField(1, required=True)
  upgradeClusterRequest = _messages.MessageField('UpgradeClusterRequest', 2)


class EdgecontainerProjectsLocationsGetRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsGetRequest object.

  Fields:
    name: Resource name for the location.
  """

  name = _messages.StringField(1, required=True)


class EdgecontainerProjectsLocationsGetServerConfigRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsGetServerConfigRequest object.

  Fields:
    name: Required. The name (project and location) of the server config to
      get, specified in the format `projects/*/locations/*`.
  """

  name = _messages.StringField(1, required=True)


class EdgecontainerProjectsLocationsListRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsListRequest object.

  Fields:
    extraLocationTypes: Optional. Do not use this field unless explicitly
      documented otherwise. This is primarily for internal usage.
    filter: A filter to narrow down results to a preferred subset. The
      filtering language accepts strings like `"displayName=tokyo"`, and is
      documented in more detail in [AIP-160](https://google.aip.dev/160).
    includeUnrevealedLocations: If true, the returned list will include
      locations which are not yet revealed.
    name: The resource that owns the locations collection, if applicable.
    pageSize: The maximum number of results to return. If not set, the service
      selects a default.
    pageToken: A page token received from the `next_page_token` field in the
      response. Send that page token to receive the subsequent page.
  """

  extraLocationTypes = _messages.StringField(1, repeated=True)
  filter = _messages.StringField(2)
  includeUnrevealedLocations = _messages.BooleanField(3)
  name = _messages.StringField(4, required=True)
  pageSize = _messages.IntegerField(5, variant=_messages.Variant.INT32)
  pageToken = _messages.StringField(6)


class EdgecontainerProjectsLocationsMachinesGetRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsMachinesGetRequest object.

  Fields:
    name: Required. The resource name of the machine.
  """

  name = _messages.StringField(1, required=True)


class EdgecontainerProjectsLocationsMachinesListRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsMachinesListRequest object.

  Fields:
    filter: Only resources matching this filter will be listed.
    orderBy: Specifies the order in which machines will be listed. In
      accordance with AIP-132, the value should be a comma-separated list of
      fields. A " desc" may be appended to a field to specify descending
      order. Supported fields: * `name` (default) : Canonical resource name of
      the machine. Example: "name desc".
    pageSize: The maximum number of resources to list.
    pageToken: A page token received from previous list request.
    parent: Required. The parent site, which owns this collection of machines.
  """

  filter = _messages.StringField(1)
  orderBy = _messages.StringField(2)
  pageSize = _messages.IntegerField(3, variant=_messages.Variant.INT32)
  pageToken = _messages.StringField(4)
  parent = _messages.StringField(5, required=True)


class EdgecontainerProjectsLocationsOperationsCancelRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsOperationsCancelRequest object.

  Fields:
    cancelOperationRequest: A CancelOperationRequest resource to be passed as
      the request body.
    name: The name of the operation resource to be cancelled.
  """

  cancelOperationRequest = _messages.MessageField('CancelOperationRequest', 1)
  name = _messages.StringField(2, required=True)


class EdgecontainerProjectsLocationsOperationsDeleteRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsOperationsDeleteRequest object.

  Fields:
    name: The name of the operation resource to be deleted.
  """

  name = _messages.StringField(1, required=True)


class EdgecontainerProjectsLocationsOperationsGetRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsOperationsGetRequest object.

  Fields:
    name: The name of the operation resource.
  """

  name = _messages.StringField(1, required=True)


class EdgecontainerProjectsLocationsOperationsListRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsOperationsListRequest object.

  Fields:
    filter: The standard list filter.
    name: The name of the operation's parent resource.
    pageSize: The standard list page size.
    pageToken: The standard list page token.
    returnPartialSuccess: When set to `true`, operations that are reachable
      are returned as normal, and those that are unreachable are returned in
      the ListOperationsResponse.unreachable field. This can only be `true`
      when reading across collections. For example, when `parent` is set to
      `"projects/example/locations/-"`. This field is not supported by default
      and will result in an `UNIMPLEMENTED` error if set unless explicitly
      documented otherwise in service or product specific documentation.
  """

  filter = _messages.StringField(1)
  name = _messages.StringField(2, required=True)
  pageSize = _messages.IntegerField(3, variant=_messages.Variant.INT32)
  pageToken = _messages.StringField(4)
  returnPartialSuccess = _messages.BooleanField(5)


class EdgecontainerProjectsLocationsVpnConnectionsCreateRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsVpnConnectionsCreateRequest object.

  Fields:
    parent: Required. The parent location where this vpn connection will be
      created.
    requestId: A unique identifier for this request. Restricted to 36 ASCII
      characters. A random UUID is recommended. This request is only
      idempotent if `request_id` is provided.
    vpnConnection: A VpnConnection resource to be passed as the request body.
    vpnConnectionId: Required. The VPN connection identifier.
  """

  parent = _messages.StringField(1, required=True)
  requestId = _messages.StringField(2)
  vpnConnection = _messages.MessageField('VpnConnection', 3)
  vpnConnectionId = _messages.StringField(4)


class EdgecontainerProjectsLocationsVpnConnectionsDeleteRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsVpnConnectionsDeleteRequest object.

  Fields:
    name: Required. The resource name of the vpn connection.
    requestId: A unique identifier for this request. Restricted to 36 ASCII
      characters. A random UUID is recommended. This request is only
      idempotent if `request_id` is provided.
  """

  name = _messages.StringField(1, required=True)
  requestId = _messages.StringField(2)


class EdgecontainerProjectsLocationsVpnConnectionsGetRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsVpnConnectionsGetRequest object.

  Fields:
    name: Required. The resource name of the vpn connection.
  """

  name = _messages.StringField(1, required=True)


class EdgecontainerProjectsLocationsVpnConnectionsListRequest(_messages.Message):
  r"""A EdgecontainerProjectsLocationsVpnConnectionsListRequest object.

  Fields:
    filter: Only resources matching this filter will be listed.
    orderBy: Specifies the order in which resources will be listed.
    pageSize: The maximum number of resources to list.
    pageToken: A page token received from previous list request.
    parent: Required. The parent location, which owns this collection of VPN
      connections.
  """

  filter = _messages.StringField(1)
  orderBy = _messages.StringField(2)
  pageSize = _messages.IntegerField(3, variant=_messages.Variant.INT32)
  pageToken = _messages.StringField(4)
  parent = _messages.StringField(5, required=True)


class Empty(_messages.Message):
  r"""A generic empty message that you can re-use to avoid defining duplicated
  empty messages in your APIs. A typical example is to use it as the request
  or the response type of an API method. For instance: service Foo { rpc
  Bar(google.protobuf.Empty) returns (google.protobuf.Empty); }
  """



class ExternalLoadBalancerPool(_messages.Message):
  r"""External load balancer pool with custom config such as name, manual/auto
  assign, non-overlapping ipv4 and optional ipv6 address range.

  Fields:
    addressPool: Optional. Name of the external load balancer pool.
    avoidBuggyIps: Optional. If true, the pool omits IP addresses ending in .0
      and .255. Some network hardware drops traffic to these special
      addresses. Its default value is false.
    ipv4Range: Required. Non-overlapping IPv4 address range of the external
      load balancer pool.
    ipv6Range: Optional. Non-overlapping IPv6 address range of the external
      load balancer pool.
    manualAssign: Optional. If true, addresses in this pool are not
      automatically assigned to Kubernetes Services. If true, an IP address in
      this pool is used only when it is specified explicitly by a service. Its
      default value is false.
  """

  addressPool = _messages.StringField(1)
  avoidBuggyIps = _messages.BooleanField(2)
  ipv4Range = _messages.StringField(3, repeated=True)
  ipv6Range = _messages.StringField(4, repeated=True)
  manualAssign = _messages.BooleanField(5)


class Fleet(_messages.Message):
  r"""Fleet related configuration. Fleets are a Google Cloud concept for
  logically organizing clusters, letting you use and manage multi-cluster
  capabilities and apply consistent policies across your systems.

  Fields:
    membership: Output only. The name of the managed Hub Membership resource
      associated to this cluster. Membership names are formatted as
      `projects//locations/global/membership/`.
    project: Required. The name of the Fleet host project where this cluster
      will be registered. Project names are formatted as `projects/`.
  """

  membership = _messages.StringField(1)
  project = _messages.StringField(2)


class GenerateAccessTokenResponse(_messages.Message):
  r"""An access token for a cluster.

  Fields:
    accessToken: Output only. Access token to authenticate to k8s api-server.
    expireTime: Output only. Timestamp at which the token will expire.
  """

  accessToken = _messages.StringField(1)
  expireTime = _messages.StringField(2)


class GenerateOfflineCredentialResponse(_messages.Message):
  r"""An offline credential for a cluster.

  Fields:
    clientCertificate: Output only. Client certificate to authenticate to k8s
      api-server.
    clientKey: Output only. Client private key to authenticate to k8s api-
      server.
    expireTime: Output only. Timestamp at which this credential will expire.
    userId: Output only. Client's identity.
  """

  clientCertificate = _messages.StringField(1)
  clientKey = _messages.StringField(2)
  expireTime = _messages.StringField(3)
  userId = _messages.StringField(4)


class GoogleGroupAuthenticationConfig(_messages.Message):
  r"""Google Group authentication config of the cluster. go/gdc-google-group-
  authentication

  Fields:
    enable: Optional. If true, the cluster will be configured to use Google
      Group authentication.
  """

  enable = _messages.BooleanField(1)


class Ingress(_messages.Message):
  r"""Config for the Ingress add-on which allows customers to create an
  Ingress object to manage external access to the servers in a cluster. The
  add-on consists of istiod and istio-ingress.

  Fields:
    disabled: Optional. Whether Ingress is disabled.
    ipv4Vip: Optional. Ingress VIP.
  """

  disabled = _messages.BooleanField(1)
  ipv4Vip = _messages.StringField(2)


class ListClustersResponse(_messages.Message):
  r"""List of clusters in a location.

  Fields:
    clusters: Clusters in the location.
    nextPageToken: A token to retrieve next page of results.
    unreachable: Locations that could not be reached.
  """

  clusters = _messages.MessageField('Cluster', 1, repeated=True)
  nextPageToken = _messages.StringField(2)
  unreachable = _messages.StringField(3, repeated=True)


class ListLocationsResponse(_messages.Message):
  r"""The response message for Locations.ListLocations.

  Fields:
    locations: A list of locations that matches the specified filter in the
      request.
    nextPageToken: The standard List next-page token.
  """

  locations = _messages.MessageField('Location', 1, repeated=True)
  nextPageToken = _messages.StringField(2)


class ListMachinesResponse(_messages.Message):
  r"""List of machines in a site.

  Fields:
    machines: Machines in the site.
    nextPageToken: A token to retrieve next page of results.
    unreachable: Locations that could not be reached.
  """

  machines = _messages.MessageField('Machine', 1, repeated=True)
  nextPageToken = _messages.StringField(2)
  unreachable = _messages.StringField(3, repeated=True)


class ListNodePoolsResponse(_messages.Message):
  r"""List of node pools in a cluster.

  Fields:
    nextPageToken: A token to retrieve next page of results.
    nodePools: Node pools in the cluster.
    unreachable: Locations that could not be reached.
  """

  nextPageToken = _messages.StringField(1)
  nodePools = _messages.MessageField('NodePool', 2, repeated=True)
  unreachable = _messages.StringField(3, repeated=True)


class ListOperationsResponse(_messages.Message):
  r"""The response message for Operations.ListOperations.

  Fields:
    nextPageToken: The standard List next-page token.
    operations: A list of operations that matches the specified filter in the
      request.
    unreachable: Unordered list. Unreachable resources. Populated when the
      request sets `ListOperationsRequest.return_partial_success` and reads
      across collections. For example, when attempting to list all resources
      across all supported locations.
  """

  nextPageToken = _messages.StringField(1)
  operations = _messages.MessageField('Operation', 2, repeated=True)
  unreachable = _messages.StringField(3, repeated=True)


class ListVpnConnectionsResponse(_messages.Message):
  r"""List of VPN connections in a location.

  Fields:
    nextPageToken: A token to retrieve next page of results.
    unreachable: Locations that could not be reached.
    vpnConnections: VpnConnections in the location.
  """

  nextPageToken = _messages.StringField(1)
  unreachable = _messages.StringField(2, repeated=True)
  vpnConnections = _messages.MessageField('VpnConnection', 3, repeated=True)


class Local(_messages.Message):
  r"""Configuration specific to clusters with a control plane hosted locally.
  Warning: Local control plane clusters must be created in their own project.
  Local control plane clusters cannot coexist in the same project with any
  other type of clusters, including non-GDCE clusters. Mixing local control
  plane GDCE clusters with any other type of clusters in the same project can
  result in data loss.

  Enums:
    ControlPlaneNodeSystemPartitionSizeValueValuesEnum: Optional. System
      partition size for control plane nodes in GiB.
    SharedDeploymentPolicyValueValuesEnum: Policy configuration about how user
      applications are deployed.

  Fields:
    controlPlaneNodeStorageSchema: Optional. Name for the storage schema of
      control plane nodes.
    controlPlaneNodeSystemPartitionSize: Optional. System partition size for
      control plane nodes in GiB.
    machineFilter: Only machines matching this filter will be allowed to host
      control plane nodes. The filtering language accepts strings like
      "name=", and is documented here: [AIP-160](https://google.aip.dev/160).
    nodeCount: The number of nodes to serve as replicas of the Control Plane.
    nodeLocation: Name of the Google Distributed Cloud Edge zones where this
      node pool will be created. For example: `us-central1-edge-customer-a`.
    sharedDeploymentPolicy: Policy configuration about how user applications
      are deployed.
  """

  class ControlPlaneNodeSystemPartitionSizeValueValuesEnum(_messages.Enum):
    r"""Optional. System partition size for control plane nodes in GiB.

    Values:
      SYSTEM_PARTITION_GIB_SIZE_UNSPECIFIED: System partition size is not
        specified by the user. This will cause the system partition size to be
        set to the size specified in the system storage schema applicable to
        that node, which is 100GiB.
      SYSTEM_PARTITION_GIB_SIZE100: 100GiB system partition size, also the
        default size.
      SYSTEM_PARTITION_GIB_SIZE300: 300GiB system partition size. Can be used
        to override the 100GiB default.
    """
    SYSTEM_PARTITION_GIB_SIZE_UNSPECIFIED = 0
    SYSTEM_PARTITION_GIB_SIZE100 = 1
    SYSTEM_PARTITION_GIB_SIZE300 = 2

  class SharedDeploymentPolicyValueValuesEnum(_messages.Enum):
    r"""Policy configuration about how user applications are deployed.

    Values:
      SHARED_DEPLOYMENT_POLICY_UNSPECIFIED: Unspecified.
      ALLOWED: User applications can be deployed both on control plane and
        worker nodes.
      DISALLOWED: User applications can not be deployed on control plane nodes
        and can only be deployed on worker nodes.
    """
    SHARED_DEPLOYMENT_POLICY_UNSPECIFIED = 0
    ALLOWED = 1
    DISALLOWED = 2

  controlPlaneNodeStorageSchema = _messages.StringField(1)
  controlPlaneNodeSystemPartitionSize = _messages.EnumField('ControlPlaneNodeSystemPartitionSizeValueValuesEnum', 2)
  machineFilter = _messages.StringField(3)
  nodeCount = _messages.IntegerField(4, variant=_messages.Variant.INT32)
  nodeLocation = _messages.StringField(5)
  sharedDeploymentPolicy = _messages.EnumField('SharedDeploymentPolicyValueValuesEnum', 6)


class LocalDiskEncryption(_messages.Message):
  r"""Configuration for CMEK support for edge machine local disk encryption.

  Enums:
    KmsKeyStateValueValuesEnum: Output only. Availability of the Cloud KMS
      CryptoKey. If not `KEY_AVAILABLE`, then nodes may go offline as they
      cannot access their local data. This can be caused by a lack of
      permissions to use the key, or if the key is disabled or deleted.
    ResourceStateValueValuesEnum: Output only. The current resource state
      associated with the cmek.

  Fields:
    kmsKey: Optional. The Cloud KMS CryptoKey e.g. projects/{project}/location
      s/{location}/keyRings/{keyRing}/cryptoKeys/{cryptoKey} to use for
      protecting node local disks. If not specified, a Google-managed key will
      be used instead.
    kmsKeyActiveVersion: Output only. The Cloud KMS CryptoKeyVersion currently
      in use for protecting node local disks. Only applicable if kms_key is
      set.
    kmsKeyState: Output only. Availability of the Cloud KMS CryptoKey. If not
      `KEY_AVAILABLE`, then nodes may go offline as they cannot access their
      local data. This can be caused by a lack of permissions to use the key,
      or if the key is disabled or deleted.
    kmsStatus: Output only. Error status returned by Cloud KMS when using this
      key. This field may be populated only if `kms_key_state` is not
      `KMS_KEY_STATE_KEY_AVAILABLE`. If populated, this field contains the
      error status reported by Cloud KMS.
    resourceState: Output only. The current resource state associated with the
      cmek.
  """

  class KmsKeyStateValueValuesEnum(_messages.Enum):
    r"""Output only. Availability of the Cloud KMS CryptoKey. If not
    `KEY_AVAILABLE`, then nodes may go offline as they cannot access their
    local data. This can be caused by a lack of permissions to use the key, or
    if the key is disabled or deleted.

    Values:
      KMS_KEY_STATE_UNSPECIFIED: Unspecified.
      KMS_KEY_STATE_KEY_AVAILABLE: The key is available for use, and dependent
        resources should be accessible.
      KMS_KEY_STATE_KEY_UNAVAILABLE: The key is unavailable for an unspecified
        reason. Dependent resources may be inaccessible.
    """
    KMS_KEY_STATE_UNSPECIFIED = 0
    KMS_KEY_STATE_KEY_AVAILABLE = 1
    KMS_KEY_STATE_KEY_UNAVAILABLE = 2

  class ResourceStateValueValuesEnum(_messages.Enum):
    r"""Output only. The current resource state associated with the cmek.

    Values:
      RESOURCE_STATE_UNSPECIFIED: Default value.
      RESOURCE_STATE_LOCK_DOWN: The resource is in LOCK DOWN state.
      RESOURCE_STATE_LOCK_DOWN_PENDING: The resource is pending lock down.
    """
    RESOURCE_STATE_UNSPECIFIED = 0
    RESOURCE_STATE_LOCK_DOWN = 1
    RESOURCE_STATE_LOCK_DOWN_PENDING = 2

  kmsKey = _messages.StringField(1)
  kmsKeyActiveVersion = _messages.StringField(2)
  kmsKeyState = _messages.EnumField('KmsKeyStateValueValuesEnum', 3)
  kmsStatus = _messages.MessageField('Status', 4)
  resourceState = _messages.EnumField('ResourceStateValueValuesEnum', 5)


class Location(_messages.Message):
  r"""A resource that represents a Google Cloud location.

  Messages:
    LabelsValue: Cross-service attributes for the location. For example
      {"cloud.googleapis.com/region": "us-east1"}
    MetadataValue: Service-specific metadata. For example the available
      capacity at the given location.

  Fields:
    displayName: The friendly name for this location, typically a nearby city
      name. For example, "Tokyo".
    labels: Cross-service attributes for the location. For example
      {"cloud.googleapis.com/region": "us-east1"}
    locationId: The canonical id for this location. For example: `"us-east1"`.
    metadata: Service-specific metadata. For example the available capacity at
      the given location.
    name: Resource name for the location, which may vary between
      implementations. For example: `"projects/example-project/locations/us-
      east1"`
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class LabelsValue(_messages.Message):
    r"""Cross-service attributes for the location. For example
    {"cloud.googleapis.com/region": "us-east1"}

    Messages:
      AdditionalProperty: An additional property for a LabelsValue object.

    Fields:
      additionalProperties: Additional properties of type LabelsValue
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a LabelsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  @encoding.MapUnrecognizedFields('additionalProperties')
  class MetadataValue(_messages.Message):
    r"""Service-specific metadata. For example the available capacity at the
    given location.

    Messages:
      AdditionalProperty: An additional property for a MetadataValue object.

    Fields:
      additionalProperties: Properties of the object. Contains field @type
        with type URL.
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a MetadataValue object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  displayName = _messages.StringField(1)
  labels = _messages.MessageField('LabelsValue', 2)
  locationId = _messages.StringField(3)
  metadata = _messages.MessageField('MetadataValue', 4)
  name = _messages.StringField(5)


class LocationMetadata(_messages.Message):
  r"""Metadata for a given google.cloud.location.Location.

  Messages:
    AvailableZonesValue: The set of available Google Distributed Cloud Edge
      zones in the location. The map is keyed by the lowercase ID of each
      zone.

  Fields:
    availableZones: The set of available Google Distributed Cloud Edge zones
      in the location. The map is keyed by the lowercase ID of each zone.
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class AvailableZonesValue(_messages.Message):
    r"""The set of available Google Distributed Cloud Edge zones in the
    location. The map is keyed by the lowercase ID of each zone.

    Messages:
      AdditionalProperty: An additional property for a AvailableZonesValue
        object.

    Fields:
      additionalProperties: Additional properties of type AvailableZonesValue
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a AvailableZonesValue object.

      Fields:
        key: Name of the additional property.
        value: A ZoneMetadata attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('ZoneMetadata', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  availableZones = _messages.MessageField('AvailableZonesValue', 1)


class Machine(_messages.Message):
  r"""A Google Distributed Cloud Edge machine capable of acting as a
  Kubernetes node.

  Enums:
    PurposeValueValuesEnum: The type of cluster the machine is used for.
    StatusValueValuesEnum: Output only. The current status of the machine.

  Messages:
    LabelsValue: Labels associated with this resource.

  Fields:
    createTime: Output only. The time when the node pool was created.
    disabled: Output only. Whether the machine is disabled. If disabled, the
      machine is unable to enter service.
    hostedNode: Canonical resource name of the node that this machine is
      responsible for hosting e.g. projects/{project}/locations/{location}/clu
      sters/{cluster_id}/nodePools/{pool_id}/{node}, Or empty if the machine
      is not assigned to assume the role of a node. For control plane nodes
      hosted on edge machines, this will return the following format: "project
      s/{project}/locations/{location}/clusters/{cluster_id}/controlPlaneNodes
      /{node}".
    labels: Labels associated with this resource.
    name: Required. The resource name of the machine.
    purpose: The type of cluster the machine is used for.
    status: Output only. The current status of the machine.
    updateTime: Output only. The time when the node pool was last updated.
    version: Output only. The software version of the machine.
    zone: The Google Distributed Cloud Edge zone of this machine.
  """

  class PurposeValueValuesEnum(_messages.Enum):
    r"""The type of cluster the machine is used for.

    Values:
      PURPOSE_UNSPECIFIED: Unspecified purpose.
      VIRTUALIZED_WORKLOAD: Machine is used for virtual workload.
      BAREMETAL_CLUSTER: Machine is used for a baremetal user cluster.
    """
    PURPOSE_UNSPECIFIED = 0
    VIRTUALIZED_WORKLOAD = 1
    BAREMETAL_CLUSTER = 2

  class StatusValueValuesEnum(_messages.Enum):
    r"""Output only. The current status of the machine.

    Values:
      STATUS_UNSPECIFIED: Status unknown.
      READY: The machine is ready to host a node. This is the default.
      DISABLED_FOR_REPAIR: The machine has been disabled for repair by adding
        1 or more disable claims.
    """
    STATUS_UNSPECIFIED = 0
    READY = 1
    DISABLED_FOR_REPAIR = 2

  @encoding.MapUnrecognizedFields('additionalProperties')
  class LabelsValue(_messages.Message):
    r"""Labels associated with this resource.

    Messages:
      AdditionalProperty: An additional property for a LabelsValue object.

    Fields:
      additionalProperties: Additional properties of type LabelsValue
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a LabelsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  createTime = _messages.StringField(1)
  disabled = _messages.BooleanField(2)
  hostedNode = _messages.StringField(3)
  labels = _messages.MessageField('LabelsValue', 4)
  name = _messages.StringField(5)
  purpose = _messages.EnumField('PurposeValueValuesEnum', 6)
  status = _messages.EnumField('StatusValueValuesEnum', 7)
  updateTime = _messages.StringField(8)
  version = _messages.StringField(9)
  zone = _messages.StringField(10)


class MaintenanceEvent(_messages.Message):
  r"""A Maintenance Event is an operation that could cause temporary
  disruptions to the cluster workloads, including Google-driven or user-
  initiated cluster upgrades, user-initiated cluster configuration changes
  that require restarting nodes, etc.

  Enums:
    ScheduleValueValuesEnum: Output only. The schedule of the maintenance
      event.
    StateValueValuesEnum: Output only. The state of the maintenance event.
    TypeValueValuesEnum: Output only. The type of the maintenance event.

  Fields:
    createTime: Output only. The time when the maintenance event request was
      created.
    endTime: Output only. The time when the maintenance event ended, either
      successfully or not. If the maintenance event is split into multiple
      maintenance windows, end_time is only updated when the whole flow ends.
    operation: Output only. The operation for running the maintenance event.
      Specified in the format projects/*/locations/*/operations/*. If the
      maintenance event is split into multiple operations (e.g. due to
      maintenance windows), the latest one is recorded.
    schedule: Output only. The schedule of the maintenance event.
    startTime: Output only. The time when the maintenance event started.
    state: Output only. The state of the maintenance event.
    targetVersion: Output only. The target version of the cluster.
    type: Output only. The type of the maintenance event.
    updateTime: Output only. The time when the maintenance event message was
      updated.
    uuid: Output only. UUID of the maintenance event.
  """

  class ScheduleValueValuesEnum(_messages.Enum):
    r"""Output only. The schedule of the maintenance event.

    Values:
      SCHEDULE_UNSPECIFIED: Unspecified.
      IMMEDIATELY: Immediately after receiving the request.
    """
    SCHEDULE_UNSPECIFIED = 0
    IMMEDIATELY = 1

  class StateValueValuesEnum(_messages.Enum):
    r"""Output only. The state of the maintenance event.

    Values:
      STATE_UNSPECIFIED: Unspecified.
      RECONCILING: The maintenance event is ongoing. The cluster might be
        unusable.
      SUCCEEDED: The maintenance event succeeded.
      FAILED: The maintenance event failed.
      STOPPED_BEFORE_MAINTENANCE_WINDOW_ENDED: The maintenance event is
        paused. The cluster should be usable.
    """
    STATE_UNSPECIFIED = 0
    RECONCILING = 1
    SUCCEEDED = 2
    FAILED = 3
    STOPPED_BEFORE_MAINTENANCE_WINDOW_ENDED = 4

  class TypeValueValuesEnum(_messages.Enum):
    r"""Output only. The type of the maintenance event.

    Values:
      TYPE_UNSPECIFIED: Unspecified.
      USER_INITIATED_UPGRADE: Upgrade initiated by users.
      GOOGLE_DRIVEN_UPGRADE: Upgrade driven by Google.
    """
    TYPE_UNSPECIFIED = 0
    USER_INITIATED_UPGRADE = 1
    GOOGLE_DRIVEN_UPGRADE = 2

  createTime = _messages.StringField(1)
  endTime = _messages.StringField(2)
  operation = _messages.StringField(3)
  schedule = _messages.EnumField('ScheduleValueValuesEnum', 4)
  startTime = _messages.StringField(5)
  state = _messages.EnumField('StateValueValuesEnum', 6)
  targetVersion = _messages.StringField(7)
  type = _messages.EnumField('TypeValueValuesEnum', 8)
  updateTime = _messages.StringField(9)
  uuid = _messages.StringField(10)


class MaintenanceExclusionWindow(_messages.Message):
  r"""Represents a maintenance exclusion window.

  Fields:
    id: Optional. A unique (per cluster) id for the window.
    window: Optional. The time window.
  """

  id = _messages.StringField(1)
  window = _messages.MessageField('TimeWindow', 2)


class MaintenancePolicy(_messages.Message):
  r"""Maintenance policy configuration.

  Fields:
    maintenanceExclusions: Optional. Exclusions to automatic maintenance. Non-
      emergency maintenance should not occur in these windows. Each exclusion
      has a unique name and may be active or expired. The max number of
      maintenance exclusions allowed at a given time is 3.
    window: Specifies the maintenance window in which maintenance may be
      performed.
  """

  maintenanceExclusions = _messages.MessageField('MaintenanceExclusionWindow', 1, repeated=True)
  window = _messages.MessageField('MaintenanceWindow', 2)


class MaintenanceWindow(_messages.Message):
  r"""Maintenance window configuration

  Fields:
    recurringWindow: Configuration of a recurring maintenance window.
  """

  recurringWindow = _messages.MessageField('RecurringTimeWindow', 1)


class NodeConfig(_messages.Message):
  r"""Configuration for each node in the NodePool

  Enums:
    NodeSystemPartitionSizeValueValuesEnum: Optional. System partition size
      for worker nodes in GiB.

  Messages:
    LabelsValue: Optional. The Kubernetes node labels

  Fields:
    labels: Optional. The Kubernetes node labels
    nodeStorageSchema: Optional. Name for the storage schema of worker nodes.
    nodeSystemPartitionSize: Optional. System partition size for worker nodes
      in GiB.
  """

  class NodeSystemPartitionSizeValueValuesEnum(_messages.Enum):
    r"""Optional. System partition size for worker nodes in GiB.

    Values:
      SYSTEM_PARTITION_GIB_SIZE_UNSPECIFIED: System partition size is not
        specified by the user. This will cause the system partition size to be
        set to the size specified in the system storage schema applicable to
        that node, which is 100GiB.
      SYSTEM_PARTITION_GIB_SIZE100: 100GiB system partition size, also the
        default size.
      SYSTEM_PARTITION_GIB_SIZE300: 300GiB system partition size. Can be used
        to override the 100GiB default.
    """
    SYSTEM_PARTITION_GIB_SIZE_UNSPECIFIED = 0
    SYSTEM_PARTITION_GIB_SIZE100 = 1
    SYSTEM_PARTITION_GIB_SIZE300 = 2

  @encoding.MapUnrecognizedFields('additionalProperties')
  class LabelsValue(_messages.Message):
    r"""Optional. The Kubernetes node labels

    Messages:
      AdditionalProperty: An additional property for a LabelsValue object.

    Fields:
      additionalProperties: Additional properties of type LabelsValue
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a LabelsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  labels = _messages.MessageField('LabelsValue', 1)
  nodeStorageSchema = _messages.StringField(2)
  nodeSystemPartitionSize = _messages.EnumField('NodeSystemPartitionSizeValueValuesEnum', 3)


class NodePool(_messages.Message):
  r"""A set of Kubernetes nodes in a cluster with common configuration and
  specification.

  Messages:
    LabelsValue: Labels associated with this resource.

  Fields:
    createTime: Output only. The time when the node pool was created.
    labels: Labels associated with this resource.
    localDiskEncryption: Optional. Local disk encryption options. This field
      is only used when enabling CMEK support.
    machineFilter: Only machines matching this filter will be allowed to join
      the node pool. The filtering language accepts strings like "name=", and
      is documented in more detail in [AIP-160](https://google.aip.dev/160).
    name: Required. The resource name of the node pool.
    nodeConfig: Optional. Configuration for each node in the NodePool
    nodeCount: Required. The number of nodes in the pool.
    nodeLocation: Name of the Google Distributed Cloud Edge zone where this
      node pool will be created. For example: `us-central1-edge-customer-a`.
    nodeVersion: Output only. The lowest release version among all worker
      nodes.
    site: Name of the Google Distributed Cloud Edge zone where this node pool
      will be created.
    updateTime: Output only. The time when the node pool was last updated.
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class LabelsValue(_messages.Message):
    r"""Labels associated with this resource.

    Messages:
      AdditionalProperty: An additional property for a LabelsValue object.

    Fields:
      additionalProperties: Additional properties of type LabelsValue
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a LabelsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  createTime = _messages.StringField(1)
  labels = _messages.MessageField('LabelsValue', 2)
  localDiskEncryption = _messages.MessageField('LocalDiskEncryption', 3)
  machineFilter = _messages.StringField(4)
  name = _messages.StringField(5)
  nodeConfig = _messages.MessageField('NodeConfig', 6)
  nodeCount = _messages.IntegerField(7, variant=_messages.Variant.INT32)
  nodeLocation = _messages.StringField(8)
  nodeVersion = _messages.StringField(9)
  site = _messages.StringField(10)
  updateTime = _messages.StringField(11)


class Operation(_messages.Message):
  r"""This resource represents a long-running operation that is the result of
  a network API call.

  Messages:
    MetadataValue: Service-specific metadata associated with the operation. It
      typically contains progress information and common metadata such as
      create time. Some services might not provide such metadata. Any method
      that returns a long-running operation should document the metadata type,
      if any.
    ResponseValue: The normal, successful response of the operation. If the
      original method returns no data on success, such as `Delete`, the
      response is `google.protobuf.Empty`. If the original method is standard
      `Get`/`Create`/`Update`, the response should be the resource. For other
      methods, the response should have the type `XxxResponse`, where `Xxx` is
      the original method name. For example, if the original method name is
      `TakeSnapshot()`, the inferred response type is `TakeSnapshotResponse`.

  Fields:
    done: If the value is `false`, it means the operation is still in
      progress. If `true`, the operation is completed, and either `error` or
      `response` is available.
    error: The error result of the operation in case of failure or
      cancellation.
    metadata: Service-specific metadata associated with the operation. It
      typically contains progress information and common metadata such as
      create time. Some services might not provide such metadata. Any method
      that returns a long-running operation should document the metadata type,
      if any.
    name: The server-assigned name, which is only unique within the same
      service that originally returns it. If you use the default HTTP mapping,
      the `name` should be a resource name ending with
      `operations/{unique_id}`.
    response: The normal, successful response of the operation. If the
      original method returns no data on success, such as `Delete`, the
      response is `google.protobuf.Empty`. If the original method is standard
      `Get`/`Create`/`Update`, the response should be the resource. For other
      methods, the response should have the type `XxxResponse`, where `Xxx` is
      the original method name. For example, if the original method name is
      `TakeSnapshot()`, the inferred response type is `TakeSnapshotResponse`.
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class MetadataValue(_messages.Message):
    r"""Service-specific metadata associated with the operation. It typically
    contains progress information and common metadata such as create time.
    Some services might not provide such metadata. Any method that returns a
    long-running operation should document the metadata type, if any.

    Messages:
      AdditionalProperty: An additional property for a MetadataValue object.

    Fields:
      additionalProperties: Properties of the object. Contains field @type
        with type URL.
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a MetadataValue object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  @encoding.MapUnrecognizedFields('additionalProperties')
  class ResponseValue(_messages.Message):
    r"""The normal, successful response of the operation. If the original
    method returns no data on success, such as `Delete`, the response is
    `google.protobuf.Empty`. If the original method is standard
    `Get`/`Create`/`Update`, the response should be the resource. For other
    methods, the response should have the type `XxxResponse`, where `Xxx` is
    the original method name. For example, if the original method name is
    `TakeSnapshot()`, the inferred response type is `TakeSnapshotResponse`.

    Messages:
      AdditionalProperty: An additional property for a ResponseValue object.

    Fields:
      additionalProperties: Properties of the object. Contains field @type
        with type URL.
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a ResponseValue object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  done = _messages.BooleanField(1)
  error = _messages.MessageField('Status', 2)
  metadata = _messages.MessageField('MetadataValue', 3)
  name = _messages.StringField(4)
  response = _messages.MessageField('ResponseValue', 5)


class OperationMetadata(_messages.Message):
  r"""Long-running operation metadata for Edge Container API methods.

  Enums:
    StatusReasonValueValuesEnum: Machine-readable status of the operation, if
      any.

  Fields:
    apiVersion: API version used to start the operation.
    createTime: The time the operation was created.
    endTime: The time the operation finished running.
    publicError: A public-facing error representation, intended for client
      consumption. This field is populated from the operation's error, and is
      especially useful for providing richer error details to clients when the
      canonical error code is `INTERNAL` and the primary error message may be
      sanitized.
    requestedCancellation: Identifies whether the user has requested
      cancellation of the operation. Operations that have successfully been
      cancelled have Operation.error value with a google.rpc.Status.code of 1,
      corresponding to `Code.CANCELLED`.
    statusMessage: Human-readable status of the operation, if any.
    statusReason: Machine-readable status of the operation, if any.
    target: Server-defined resource path for the target of the operation.
    verb: The verb executed by the operation.
    warnings: Warnings that do not block the operation, but still hold
      relevant information for the end user to receive.
  """

  class StatusReasonValueValuesEnum(_messages.Enum):
    r"""Machine-readable status of the operation, if any.

    Values:
      STATUS_REASON_UNSPECIFIED: Reason unknown.
      UPGRADE_PAUSED: The cluster upgrade is currently paused.
      RETRYABLE_ERROR: The request has errored, but the error is retryable.
    """
    STATUS_REASON_UNSPECIFIED = 0
    UPGRADE_PAUSED = 1
    RETRYABLE_ERROR = 2

  apiVersion = _messages.StringField(1)
  createTime = _messages.StringField(2)
  endTime = _messages.StringField(3)
  publicError = _messages.MessageField('Status', 4)
  requestedCancellation = _messages.BooleanField(5)
  statusMessage = _messages.StringField(6)
  statusReason = _messages.EnumField('StatusReasonValueValuesEnum', 7)
  target = _messages.StringField(8)
  verb = _messages.StringField(9)
  warnings = _messages.StringField(10, repeated=True)


class Quota(_messages.Message):
  r"""Represents quota for Edge Container resources.

  Fields:
    limit: Quota limit for this metric.
    metric: Name of the quota metric.
    usage: Current usage of this metric.
  """

  limit = _messages.FloatField(1)
  metric = _messages.StringField(2)
  usage = _messages.FloatField(3)


class RecurringTimeWindow(_messages.Message):
  r"""Represents an arbitrary window of time that recurs.

  Fields:
    recurrence: An RRULE (https://tools.ietf.org/html/rfc5545#section-3.8.5.3)
      for how this window recurs. They go on for the span of time between the
      start and end time.
    window: The window of the first recurrence.
  """

  recurrence = _messages.StringField(1)
  window = _messages.MessageField('TimeWindow', 2)


class Remote(_messages.Message):
  r"""Configuration specific to clusters with a control plane hosted remotely.
  """



class RobinCloudNativeStorage(_messages.Message):
  r"""Defines configurations for Robin Cloud Native Storage.

  Fields:
    enable: Optional. If true, enable Robin CNS in this cluster.
  """

  enable = _messages.BooleanField(1)


class SdsOperator(_messages.Message):
  r"""Config for the SDS Operator add-on which installs Robin CNS.

  Fields:
    version: Optional. SDS Operator version.
  """

  version = _messages.StringField(1)


class ServerConfig(_messages.Message):
  r"""Server configuration for supported versions and release channels.

  Messages:
    ChannelsValue: Output only. Mapping from release channel to channel
      config.

  Fields:
    channels: Output only. Mapping from release channel to channel config.
    defaultVersion: Output only. Default version, e.g.: "1.4.0".
    versionRollouts: Output only. Rollout information for the server config.
    versions: Output only. Supported versions, e.g.: ["1.4.0", "1.5.0"].
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class ChannelsValue(_messages.Message):
    r"""Output only. Mapping from release channel to channel config.

    Messages:
      AdditionalProperty: An additional property for a ChannelsValue object.

    Fields:
      additionalProperties: Additional properties of type ChannelsValue
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a ChannelsValue object.

      Fields:
        key: Name of the additional property.
        value: A ChannelConfig attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('ChannelConfig', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  channels = _messages.MessageField('ChannelsValue', 1)
  defaultVersion = _messages.StringField(2)
  versionRollouts = _messages.MessageField('VersionRollout', 3, repeated=True)
  versions = _messages.MessageField('Version', 4, repeated=True)


class StandardQueryParameters(_messages.Message):
  r"""Query parameters accepted by all methods.

  Enums:
    FXgafvValueValuesEnum: V1 error format.
    AltValueValuesEnum: Data format for response.

  Fields:
    f__xgafv: V1 error format.
    access_token: OAuth access token.
    alt: Data format for response.
    callback: JSONP
    fields: Selector specifying which fields to include in a partial response.
    key: API key. Your API key identifies your project and provides you with
      API access, quota, and reports. Required unless you provide an OAuth 2.0
      token.
    oauth_token: OAuth 2.0 token for the current user.
    prettyPrint: Returns response with indentations and line breaks.
    quotaUser: Available to use for quota purposes for server-side
      applications. Can be any arbitrary string assigned to a user, but should
      not exceed 40 characters.
    trace: A tracing token of the form "token:<tokenid>" to include in api
      requests.
    uploadType: Legacy upload protocol for media (e.g. "media", "multipart").
    upload_protocol: Upload protocol for media (e.g. "raw", "multipart").
  """

  class AltValueValuesEnum(_messages.Enum):
    r"""Data format for response.

    Values:
      json: Responses with Content-Type of application/json
      media: Media download with context-dependent Content-Type
      proto: Responses with Content-Type of application/x-protobuf
    """
    json = 0
    media = 1
    proto = 2

  class FXgafvValueValuesEnum(_messages.Enum):
    r"""V1 error format.

    Values:
      _1: v1 error format
      _2: v2 error format
    """
    _1 = 0
    _2 = 1

  f__xgafv = _messages.EnumField('FXgafvValueValuesEnum', 1)
  access_token = _messages.StringField(2)
  alt = _messages.EnumField('AltValueValuesEnum', 3, default='json')
  callback = _messages.StringField(4)
  fields = _messages.StringField(5)
  key = _messages.StringField(6)
  oauth_token = _messages.StringField(7)
  prettyPrint = _messages.BooleanField(8, default=True)
  quotaUser = _messages.StringField(9)
  trace = _messages.StringField(10)
  uploadType = _messages.StringField(11)
  upload_protocol = _messages.StringField(12)


class Status(_messages.Message):
  r"""The `Status` type defines a logical error model that is suitable for
  different programming environments, including REST APIs and RPC APIs. It is
  used by [gRPC](https://github.com/grpc). Each `Status` message contains
  three pieces of data: error code, error message, and error details. You can
  find out more about this error model and how to work with it in the [API
  Design Guide](https://cloud.google.com/apis/design/errors).

  Messages:
    DetailsValueListEntry: A DetailsValueListEntry object.

  Fields:
    code: The status code, which should be an enum value of google.rpc.Code.
    details: A list of messages that carry the error details. There is a
      common set of message types for APIs to use.
    message: A developer-facing error message, which should be in English. Any
      user-facing error message should be localized and sent in the
      google.rpc.Status.details field, or localized by the client.
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class DetailsValueListEntry(_messages.Message):
    r"""A DetailsValueListEntry object.

    Messages:
      AdditionalProperty: An additional property for a DetailsValueListEntry
        object.

    Fields:
      additionalProperties: Properties of the object. Contains field @type
        with type URL.
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a DetailsValueListEntry object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  code = _messages.IntegerField(1, variant=_messages.Variant.INT32)
  details = _messages.MessageField('DetailsValueListEntry', 2, repeated=True)
  message = _messages.StringField(3)


class SurvivabilityConfig(_messages.Message):
  r"""Configuration of the cluster survivability, e.g., for the case when
  network connectivity is lost.

  Fields:
    offlineRebootTtl: Optional. Time period that allows the cluster nodes to
      be rebooted and become functional without network connectivity to
      Google. The default 0 means not allowed. The maximum is 30 days.
  """

  offlineRebootTtl = _messages.StringField(1)


class SystemAddonsConfig(_messages.Message):
  r"""Config that customers are allowed to define for GDCE system add-ons.

  Fields:
    ingress: Optional. Config for Ingress.
    robinCloudNativeStorage: Optional. Configurations for Robin Cloud Native
      Storage.
    sdsOperator: Optional. Config for SDS Operator.
    unmanagedKafkaConfig: Optional. Config for unmanaged Kafka.
    vmServiceConfig: Optional. Config for VM Service.
  """

  ingress = _messages.MessageField('Ingress', 1)
  robinCloudNativeStorage = _messages.MessageField('RobinCloudNativeStorage', 2)
  sdsOperator = _messages.MessageField('SdsOperator', 3)
  unmanagedKafkaConfig = _messages.MessageField('UnmanagedKafkaConfig', 4)
  vmServiceConfig = _messages.MessageField('VMServiceConfig', 5)


class TimeWindow(_messages.Message):
  r"""Represents an arbitrary window of time.

  Fields:
    endTime: The time that the window ends. The end time must take place after
      the start time.
    startTime: The time that the window first starts.
  """

  endTime = _messages.StringField(1)
  startTime = _messages.StringField(2)


class UnmanagedKafkaConfig(_messages.Message):
  r"""Config for customer provided Kafka to receive application logs from log
  forwarder. This field is only populated for LCP clusters.

  Fields:
    brokers: Required. Comma separated string of broker addresses, with IP and
      port.
    topicKey: Optional. Kafka topic key to select a topic if multiple topics
      exist.
    topics: Required. Comma separated string of Kafka topics.
  """

  brokers = _messages.StringField(1)
  topicKey = _messages.StringField(2)
  topics = _messages.StringField(3)


class UpgradeClusterRequest(_messages.Message):
  r"""Upgrades a cluster.

  Enums:
    ScheduleValueValuesEnum: The schedule for the upgrade.

  Fields:
    requestId: A unique identifier for this request. Restricted to 36 ASCII
      characters. A random UUID is recommended. This request is only
      idempotent if `request_id` is provided.
    schedule: The schedule for the upgrade.
    targetVersion: Required. The version the cluster is going to be upgraded
      to.
  """

  class ScheduleValueValuesEnum(_messages.Enum):
    r"""The schedule for the upgrade.

    Values:
      SCHEDULE_UNSPECIFIED: Unspecified. The default is to upgrade the cluster
        immediately which is the only option today.
      IMMEDIATELY: The cluster is going to be upgraded immediately after
        receiving the request.
    """
    SCHEDULE_UNSPECIFIED = 0
    IMMEDIATELY = 1

  requestId = _messages.StringField(1)
  schedule = _messages.EnumField('ScheduleValueValuesEnum', 2)
  targetVersion = _messages.StringField(3)


class UpgradeSettings(_messages.Message):
  r"""If upgrade_settings or max_unavailable_worker_nodes is unset, the Google
  defaults will be used.

  Fields:
    maxUnavailableWorkerNodes: Optional. Maximum number of unavailable worker
      nodes during upgrade.
  """

  maxUnavailableWorkerNodes = _messages.IntegerField(1, variant=_messages.Variant.INT32)


class VMServiceConfig(_messages.Message):
  r"""VMServiceConfig defines the configuration for GDCE VM Service.

  Fields:
    vmmEnabled: Optional. Whether VMM is enabled.
  """

  vmmEnabled = _messages.BooleanField(1)


class Version(_messages.Message):
  r"""Version of a cluster.

  Fields:
    name: Output only. Name of the version, e.g.: "1.4.0".
  """

  name = _messages.StringField(1)


class VersionRollout(_messages.Message):
  r"""VersionRollout contains the Version rollout information.

  Fields:
    availableZones: Output only. List of zones in which the version has been
      rolled out, e.g.: ["us-central1", "us-west1"].
    version: Output only. Semantic version, e.g.: "1.4.0".
  """

  availableZones = _messages.StringField(1, repeated=True)
  version = _messages.StringField(2)


class VpcProject(_messages.Message):
  r"""Project detail of the VPC network.

  Fields:
    projectId: The project of the VPC to connect to. If not specified, it is
      the same as the cluster project.
    serviceAccount: Optional. Deprecated: do not use.
  """

  projectId = _messages.StringField(1)
  serviceAccount = _messages.StringField(2)


class VpnConnection(_messages.Message):
  r"""A VPN connection .

  Enums:
    BgpRoutingModeValueValuesEnum: Dynamic routing mode of the VPC network,
      `regional` or `global`.

  Messages:
    LabelsValue: Labels associated with this resource.

  Fields:
    bgpRoutingMode: Dynamic routing mode of the VPC network, `regional` or
      `global`.
    cluster: The canonical Cluster name to connect to. It is in the form of
      projects/{project}/locations/{location}/clusters/{cluster}.
    createTime: Output only. The time when the VPN connection was created.
    details: Output only. The created connection details.
    enableHighAvailability: Whether this VPN connection has HA enabled on
      cluster side. If enabled, when creating VPN connection we will attempt
      to use 2 ANG floating IPs.
    labels: Labels associated with this resource.
    name: Required. The resource name of VPN connection
    natGatewayIp: NAT gateway IP, or WAN IP address. If a customer has
      multiple NAT IPs, the customer needs to configure NAT such that only one
      external IP maps to the GMEC Anthos cluster. This is empty if NAT is not
      used.
    router: Optional. The VPN connection Cloud Router name.
    updateTime: Output only. The time when the VPN connection was last
      updated.
    vpc: The network ID of VPC to connect to.
    vpcProject: Optional. Project detail of the VPC network. Required if VPC
      is in a different project than the cluster project.
  """

  class BgpRoutingModeValueValuesEnum(_messages.Enum):
    r"""Dynamic routing mode of the VPC network, `regional` or `global`.

    Values:
      BGP_ROUTING_MODE_UNSPECIFIED: Unknown.
      REGIONAL: Regional mode.
      GLOBAL: Global mode.
    """
    BGP_ROUTING_MODE_UNSPECIFIED = 0
    REGIONAL = 1
    GLOBAL = 2

  @encoding.MapUnrecognizedFields('additionalProperties')
  class LabelsValue(_messages.Message):
    r"""Labels associated with this resource.

    Messages:
      AdditionalProperty: An additional property for a LabelsValue object.

    Fields:
      additionalProperties: Additional properties of type LabelsValue
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a LabelsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  bgpRoutingMode = _messages.EnumField('BgpRoutingModeValueValuesEnum', 1)
  cluster = _messages.StringField(2)
  createTime = _messages.StringField(3)
  details = _messages.MessageField('Details', 4)
  enableHighAvailability = _messages.BooleanField(5)
  labels = _messages.MessageField('LabelsValue', 6)
  name = _messages.StringField(7)
  natGatewayIp = _messages.StringField(8)
  router = _messages.StringField(9)
  updateTime = _messages.StringField(10)
  vpc = _messages.StringField(11)
  vpcProject = _messages.MessageField('VpcProject', 12)


class ZoneMetadata(_messages.Message):
  r"""A Google Distributed Cloud Edge zone where edge machines are located.

  Messages:
    RackTypesValue: The map keyed by rack name and has value of RackType.

  Fields:
    configData: Config data for the zone.
    quota: Quota for resources in this zone.
    rackTypes: The map keyed by rack name and has value of RackType.
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class RackTypesValue(_messages.Message):
    r"""The map keyed by rack name and has value of RackType.

    Messages:
      AdditionalProperty: An additional property for a RackTypesValue object.

    Fields:
      additionalProperties: Additional properties of type RackTypesValue
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a RackTypesValue object.

      Enums:
        ValueValueValuesEnum:

      Fields:
        key: Name of the additional property.
        value: A ValueValueValuesEnum attribute.
      """

      class ValueValueValuesEnum(_messages.Enum):
        r"""ValueValueValuesEnum enum type.

        Values:
          RACK_TYPE_UNSPECIFIED: Unspecified rack type, single rack also
            belongs to this type.
          BASE: Base rack type, a pair of two modified Config-1 racks
            containing Aggregation switches.
          EXPANSION: Expansion rack type, also known as standalone racks,
            added by customers on demand.
        """
        RACK_TYPE_UNSPECIFIED = 0
        BASE = 1
        EXPANSION = 2

      key = _messages.StringField(1)
      value = _messages.EnumField('ValueValueValuesEnum', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  configData = _messages.MessageField('ConfigData', 1)
  quota = _messages.MessageField('Quota', 2, repeated=True)
  rackTypes = _messages.MessageField('RackTypesValue', 3)


class ZoneStorageEncryption(_messages.Message):
  r"""Configuration for Zone Storage CMEK Support

  Enums:
    ResourceStateValueValuesEnum: Output only. The current resource state of
      the CMEK

  Fields:
    kmsKey: Optional. The Cloud KMS Key
    kmsKeyActiveVersion: Output only. The Cloud KMS CryptoKeyVersion currently
      used for encryption/decryption
    resourceState: Output only. The current resource state of the CMEK
  """

  class ResourceStateValueValuesEnum(_messages.Enum):
    r"""Output only. The current resource state of the CMEK

    Values:
      RESOURCE_STATE_UNSPECIFIED: Default value.
      RESOURCE_STATE_LOCK_DOWN: The resource is in LOCK DOWN state.
      RESOURCE_STATE_LOCK_DOWN_PENDING: The resource is pending lock down.
    """
    RESOURCE_STATE_UNSPECIFIED = 0
    RESOURCE_STATE_LOCK_DOWN = 1
    RESOURCE_STATE_LOCK_DOWN_PENDING = 2

  kmsKey = _messages.StringField(1)
  kmsKeyActiveVersion = _messages.StringField(2)
  resourceState = _messages.EnumField('ResourceStateValueValuesEnum', 3)


encoding.AddCustomJsonFieldMapping(
    StandardQueryParameters, 'f__xgafv', '$.xgafv')
encoding.AddCustomJsonEnumMapping(
    StandardQueryParameters.FXgafvValueValuesEnum, '_1', '1')
encoding.AddCustomJsonEnumMapping(
    StandardQueryParameters.FXgafvValueValuesEnum, '_2', '2')



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Download workflow used by GCS gRPC bidi streaming client."""

from __future__ import annotations

import binascii
import io
import queue
from typing import Any, Callable, Tuple

from googlecloudsdk.api_lib.storage import cloud_api
from googlecloudsdk.api_lib.storage import errors as cloud_errors
from googlecloudsdk.api_lib.storage import gcs_download
from googlecloudsdk.api_lib.storage.gcs_grpc import grpc_util
from googlecloudsdk.api_lib.storage.gcs_grpc import retry_util as grpc_retry_util
from googlecloudsdk.api_lib.storage.gcs_grpc_bidi_streaming import retry_util
from googlecloudsdk.command_lib.storage import errors as storage_errors
from googlecloudsdk.command_lib.storage import fast_crc32c_util
from googlecloudsdk.command_lib.storage import hash_util
from googlecloudsdk.command_lib.util import crc32c
from googlecloudsdk.core import log


class BidiDownloadIncompleteError(cloud_errors.RetryableApiError):
  """Raised when BiDi download is incomplete and should be retried."""


# read_id is hardcoded to 1 for simple downloads as we only have one range.
ONE_SHOT_READ_ID = 1


def _get_bidi_read_object_spec(
    gapic_client,
    cloud_resource,
    decryption_key,
    bucket_name,
    read_handle=None,
):
  """Returns a bidi read object spec."""
  bidi_read_object_spec = gapic_client.types.BidiReadObjectSpec(
      bucket=bucket_name,
      object_=cloud_resource.storage_url.resource_name,
      generation=(
          int(cloud_resource.generation) if cloud_resource.generation else None
      ),
      common_object_request_params=grpc_util.get_encryption_request_params(
          gapic_client, decryption_key
      ),
  )
  if read_handle:
    bidi_read_object_spec.read_handle = read_handle
  return bidi_read_object_spec


def _get_bidi_read_range(
    gapic_client, start_byte, end_byte,
):
  """Returns a bidi read range."""
  # A read_length of 0 means read until the end of the object.
  read_length = (
      end_byte - start_byte + 1
      if end_byte is not None
      else 0
  )

  return gapic_client.types.ReadRange(
      read_offset=start_byte,
      read_length=read_length,
      read_id=ONE_SHOT_READ_ID,
  )


def _get_bidi_read_object_request(
    gapic_client,
    cloud_resource,
    start_byte,
    end_byte,
    decryption_key,
    read_handle=None,
):
  """Returns a bidi read object RPC request."""
  bucket_name = grpc_util.get_full_bucket_name(
      cloud_resource.storage_url.bucket_name
  )

  read_object_spec = _get_bidi_read_object_spec(
      gapic_client, cloud_resource, decryption_key, bucket_name, read_handle
  )

  read_range = _get_bidi_read_range(gapic_client, start_byte, end_byte)

  return gapic_client.types.BidiReadObjectRequest(
      read_object_spec=read_object_spec,
      read_ranges=[read_range],
  )


def _update_digesters(digesters, data, chunk_crc32c=None):
  """Updates digesters with data."""
  if digesters is None or chunk_crc32c is None:
    return
  if hash_util.HashAlgorithm.CRC32C not in digesters or isinstance(
      digesters[hash_util.HashAlgorithm.CRC32C],
      fast_crc32c_util.DeferredCrc32c,
  ):
    return

  final_crc32c = crc32c.concat_checksums(
      int.from_bytes(digesters[hash_util.HashAlgorithm.CRC32C].digest(), 'big'),
      chunk_crc32c,
      len(data),
  )
  if isinstance(
      digesters[hash_util.HashAlgorithm.CRC32C],
      fast_crc32c_util.OnTheFlyCrc32c,
  ):
    digesters[hash_util.HashAlgorithm.CRC32C] = fast_crc32c_util.OnTheFlyCrc32c(
        final_crc32c
    )
    return
  digesters[hash_util.HashAlgorithm.CRC32C] = crc32c.get_crc32c_from_checksum(
      final_crc32c
  )


def _should_validate_chunk_integrity(digesters):
  """Returns true if chunk integrity should be validated."""
  if not digesters:
    return False
  if hash_util.HashAlgorithm.CRC32C not in digesters:
    return False
  hash_object = digesters[hash_util.HashAlgorithm.CRC32C]
  if isinstance(hash_object, fast_crc32c_util.DeferredCrc32c):
    return False
  return True


def _validate_chunk_integrity(crc32c_hash, data):
  """Validates integrity of a single chunk by comparing CRC32C hashes.

  If there is a mismatch, it raises HashMismatchError. This error
  is not retried by download retry logic, and will cause the download to fail.
  The download task runner is responsible for deleting partially downloaded
  files in case of such failures.

  Args:
    crc32c_hash (int): The server-provided CRC32C hash of the chunk.
    data (bytes): The chunk data.

  Returns:
    int: The calculated CRC32C checksum if validation is successful.
    None: If chunk validation could not be performed due to an unexpected
      exception.

  Raises:
    storage_errors.HashMismatchError: If chunk integrity check fails.
  """
  log.debug('Validating chunk integrity for CRC32C: %s', crc32c_hash)
  try:
    # Partial CRC32C calculation is done only if google_crc32c is present or if
    # streaming download is used.
    hasher = fast_crc32c_util.get_crc32c(is_streaming=True)
    hasher.update(data)
    calculated_crc32c = crc32c.get_checksum(hasher)
    if crc32c_hash != calculated_crc32c:
      expected_bytes = crc32c_hash.to_bytes(4, byteorder='big')
      calculated_bytes = calculated_crc32c.to_bytes(4, byteorder='big')
      log.error(
          'Mismatch for chunk! Expected CRC32C: %s (%s), Actual CRC32C: %s'
          ' (%s).',
          crc32c_hash,
          binascii.hexlify(expected_bytes),
          calculated_crc32c,
          binascii.hexlify(calculated_bytes),
      )
      raise storage_errors.HashMismatchError(
          'Transferred chunk CRC32C does not match.'
      )
    return calculated_crc32c
  except storage_errors.HashMismatchError:
    raise
  except Exception as e:  # pylint: disable=broad-except
    log.debug('Could not perform chunk validation for CRC32C: %s', e)
    # If we can't validate the chunk, we should continue the download without
    # failing the entire operation.
    return None


def _process_data_from_bidi_read_object_rpc(
    gapic_client,
    cloud_resource,
    download_stream,
    digesters,
    progress_callback,
    start_byte,
    end_byte,
    download_strategy,
    decryption_key,
    redirection_handler,
    read_handle=None,
):
  """Receives data from the bidi read object RPC."""
  bidi_read_object_request = _get_bidi_read_object_request(
      gapic_client,
      cloud_resource,
      start_byte,
      end_byte,
      decryption_key,
      read_handle,
  )
  bidi_read_object_rpc = (
      redirection_handler.start_bidi_rpc_with_retry_on_redirected_token_error(
          bidi_read_object_request
      )
  )
  processed_bytes = start_byte
  destination_pipe_is_broken = False
  received_read_handle = read_handle
  try:
    bidi_read_object_rpc.requests_done()

    while bidi_read_object_rpc.is_active:
      try:
        bidi_read_object_response = bidi_read_object_rpc.recv()
        if (
            bidi_read_object_response
            and bidi_read_object_response.read_handle
            and bidi_read_object_response.read_handle.handle
        ):
          received_read_handle = bidi_read_object_response.read_handle
      except (StopIteration, EOFError):
        break
      else:
        # This block executes only if the try block completes without an
        # exception.
        for object_range_data in bidi_read_object_response.object_data_ranges:
          data = object_range_data.checksummed_data.content
          if data:
            chunk_crc32c = None
            if _should_validate_chunk_integrity(digesters):
              chunk_crc32c = _validate_chunk_integrity(
                  object_range_data.checksummed_data.crc32c, data)

            try:
              download_stream.write(data)
            except BrokenPipeError:
              if download_strategy == cloud_api.DownloadStrategy.ONE_SHOT:
                log.info('Writing to download stream raised broken pipe error.')
                destination_pipe_is_broken = True
                break
              raise

            if digesters:
              _update_digesters(digesters, data, chunk_crc32c)

            processed_bytes += len(data)
            if progress_callback:
              progress_callback(processed_bytes)

        if destination_pipe_is_broken:
          break
  finally:
    # Ensures the stream is closed even if an exception is raised.
    bidi_read_object_rpc.close()
  return processed_bytes, destination_pipe_is_broken, received_read_handle


def _get_target_size(cloud_resource, start_byte, end_byte):
  """Returns the target size for the download."""
  target_size = cloud_resource.size
  if end_byte is not None:
    target_size = end_byte - start_byte + 1
  elif start_byte > 0:
    target_size = cloud_resource.size - start_byte
  return target_size


def bidi_download_object(
    gapic_client,
    cloud_resource,
    download_stream,
    digesters,
    progress_callback,
    start_byte,
    end_byte,
    download_strategy,
    decryption_key,
    redirection_handler,
):
  """Downloads the object using gRPC bidi streaming API.

  Args:
    gapic_client (StorageClient): The GAPIC API client to interact with GCS
      using gRPC.
    cloud_resource (resource_reference.ObjectResource): See
      cloud_api.CloudApi.download_object.
    download_stream (stream): Stream to send the object data to.
    digesters (dict): See cloud_api.CloudApi.download_object.
    progress_callback (function): See cloud_api.CloudApi.download_object.
    start_byte (int): Starting point for download (for resumable downloads and
      range requests). Can be set to negative to request a range of bytes
      (python equivalent of [:-3]).
    end_byte (int): Ending byte number, inclusive, for download (for range
      requests). If None, download the rest of the object.
    download_strategy (cloud_api.DownloadStrategy): Download strategy used to
      perform the download.
    decryption_key (encryption_util.EncryptionKey|None): The decryption key to
      be used to download the object if the object is encrypted.
    redirection_handler (retry_util.BidiRedirectedTokenErrorHandler): The
      redirection handler to handle redirected token errors.
  """

  target_size = _get_target_size(cloud_resource, start_byte, end_byte)
  processed_bytes, destination_pipe_is_broken = retry_util.run_with_retries(
      _process_data_from_bidi_read_object_rpc,
      gapic_client,
      cloud_resource,
      download_stream,
      digesters,
      progress_callback,
      start_byte,
      end_byte,
      download_strategy,
      decryption_key,
      target_size,
      redirection_handler,
  )

  total_downloaded_data = processed_bytes - start_byte
  if (
      target_size is not None
      and total_downloaded_data < target_size
      and not destination_pipe_is_broken
  ):
    error_message = (
        'Download not completed for %s. Target size=%s, downloaded data=%s.'
        ' The input stream terminated before the entire content was read,'
        ' possibly due to a network condition.'
    ) % (
        cloud_resource.storage_url.resource_name,
        target_size,
        total_downloaded_data,
    )
    log.debug(error_message)
    raise cloud_errors.RetryableApiError(error_message)

  return None


class BidiDownloader:
  """Helper class to manage state for resumable Bidi downloads."""

  def __init__(
      self,
      process_chunk_func: Callable[..., Tuple[int, bool]],
      gapic_client: 'storage_client_v2.StorageClient',
      cloud_resource: 'resource_reference.ObjectResource',
      download_stream: io.IOBase,
      digesters: dict[str, Any] | None,
      progress_callback: Callable[[int], None] | None,
      start_byte: int,
      end_byte: int | None,
      download_strategy: cloud_api.DownloadStrategy,
      decryption_key: 'encryption_util.EncryptionKey' | None,
      target_size: int | None,
      redirection_handler: retry_util.BidiRedirectedTokenErrorHandler,
  ):
    """Initializes a BidiDownloader instance.

    Args:
      process_chunk_func (Callable[..., Tuple[int, bool]]): Function that
        downloads a chunk of data and returns processed_bytes and
        destination_pipe_is_broken.
      gapic_client (StorageClient): The GAPIC API client to interact with GCS
        using gRPC.
      cloud_resource (resource_reference.ObjectResource): See
        cloud_api.CloudApi.download_object.
      download_stream (io.IOBase): Stream to send the object data to.
      digesters (Dict[str, 'hashlib._Hash'] | None): See
        cloud_api.CloudApi.download_object.
      progress_callback (Callable[[int], None] | None): See
        cloud_api.CloudApi.download_object.
      start_byte (int): Starting point for download.
      end_byte (int | None): Ending byte number, inclusive, for download. If
        None, download the rest of the object.
      download_strategy (cloud_api.DownloadStrategy): Download strategy used to
        perform the download.
      decryption_key (EncryptionKey | None): The decryption key to
        be used to download the object if the object is encrypted.
      target_size (int | None): The total number of bytes to download.
      redirection_handler (retry_util.BidiRedirectedTokenErrorHandler): The
        redirection handler to handle redirected token errors.
    """
    self.process_chunk_func = process_chunk_func
    self.gapic_client = gapic_client
    self.cloud_resource = cloud_resource
    self.download_stream = download_stream
    self.digesters = digesters
    self.progress_callback = progress_callback
    self.start_byte = start_byte
    self.end_byte = end_byte
    self.download_strategy = download_strategy
    self.decryption_key = decryption_key
    self.target_size = target_size
    self.processed_bytes = start_byte
    self.redirection_handler = redirection_handler
    self.destination_pipe_is_broken = False
    self.read_handle = None

  def download_chunk(self):
    """Performs one download attempt and updates processed_bytes.

    If the attempt failed with a retriable error, the download will be
    re-performed from the last processed byte.

    Raises:
      BidiDownloadIncompleteError: If the download stream ends before all bytes
        are received, triggering a retry.

    Returns:
      A tuple containing:
      - int: The total number of bytes processed.
      - bool: True if the destination pipe is broken, False otherwise.
    """
    (
        self.processed_bytes,
        self.destination_pipe_is_broken,
        self.read_handle,
    ) = self.process_chunk_func(
        self.gapic_client,
        self.cloud_resource,
        self.download_stream,
        self.digesters,
        self.progress_callback,
        self.processed_bytes,  # Resume from last processed byte.
        self.end_byte,
        self.download_strategy,
        self.decryption_key,
        self.redirection_handler,
        read_handle=self.read_handle,
    )
    total_downloaded_data = self.processed_bytes - self.start_byte
    if self.destination_pipe_is_broken:
      return self.processed_bytes, self.destination_pipe_is_broken
    if self.target_size is None or total_downloaded_data >= self.target_size:
      # Download complete.
      return self.processed_bytes, self.destination_pipe_is_broken
    raise BidiDownloadIncompleteError('Stream ended prematurely.')


class BidiGrpcDownload(gcs_download.GcsDownload):
  """Perform GCS Download using gRPC bidi streaming API."""

  def __init__(self,
               gapic_client,
               cloud_resource,
               download_stream,
               start_byte,
               end_byte,
               digesters,
               progress_callback,
               download_strategy,
               decryption_key):
    """Initializes a BidiGrpcDownload instance.

    Args:
      gapic_client (StorageClient): The GAPIC API client to interact with
        GCS using gRPC.
      cloud_resource (resource_reference.ObjectResource): See
        cloud_api.CloudApi.download_object.
      download_stream (stream): Stream to send the object data to.
      start_byte (int): See super class.
      end_byte (int): See super class.
      digesters (dict): See cloud_api.CloudApi.download_object.
      progress_callback (function): See cloud_api.CloudApi.download_object.
      download_strategy (cloud_api.DownloadStrategy): Download strategy used to
        perform the download.
      decryption_key (encryption_util.EncryptionKey|None): The decryption key to
        be used to download the object if the object is encrypted.
    """
    super(BidiGrpcDownload, self).__init__(
        download_stream, start_byte, end_byte
    )
    self._gapic_client = gapic_client
    self._cloud_resource = cloud_resource
    self._digesters = digesters
    self._progress_callback = progress_callback
    self._download_strategy = download_strategy
    self._decryption_key = decryption_key
    self._redirection_handler = retry_util.BidiRedirectedTokenErrorHandler(
        self._gapic_client,
        source_resource=self._cloud_resource,
        destination_resource=self._download_stream,
    )

  def should_retry(self, exc_type, exc_value, exc_traceback):
    """See super class."""
    return grpc_retry_util.is_retriable(exc_type, exc_value, exc_traceback)

  def launch(self):
    """See super class."""
    return bidi_download_object(
        gapic_client=self._gapic_client,
        cloud_resource=self._cloud_resource,
        download_stream=self._download_stream,
        digesters=self._digesters,
        progress_callback=self._progress_callback,
        start_byte=self._start_byte,
        end_byte=self._end_byte,
        download_strategy=self._download_strategy,
        decryption_key=self._decryption_key,
        redirection_handler=self._redirection_handler,
    )

  @grpc_retry_util.grpc_default_retryer
  def simple_download(self):
    """Downloads the object.

    On retriable errors, the entire download will be re-performed instead of
    resuming from a particular byte. This is useful for streaming download
    cases.

    Unlike Apitools, the GAPIC client doesn't retry the request by
    default, hence we are using the decorator.

    Returns:
      Encoding string for object if requested. Otherwise, None.
    """
    return self.launch()

  def run(self):
    """See super class."""
    if self._download_strategy == cloud_api.DownloadStrategy.ONE_SHOT:
      return self.simple_download()
    else:
      return super(BidiGrpcDownload, self).run(retriable_in_flight=True)


class OffsetSequencer:
  """Sequences out-of-order chunks from a multi-range download stream.

  Handles out-of-order chunks and yields contiguous blocks for the decoupled
  writer. Note that the internal buffer is theoretically unbounded; it is the
  responsibility of the client (e.g., via bounded network multiplexing)
  to ensure peak memory is mathematically capped.
  """

  def __init__(self, start_offset: int):
    self._next_expected_offset = start_offset
    self._buffer = {}  # type: dict[int, bytes]

  def add_chunk(self, offset: int, data: bytes):
    """Caches a chunk of data at the given offset.

    Args:
      offset: The byte offset where this data starts.
      data: The chunk's byte data.
    """
    if not data:
      return

    # Ignore chunks we've already processed (e.g., retried network ranges).
    if offset < self._next_expected_offset:
      overlap = self._next_expected_offset - offset
      if overlap >= len(data):
        return
      data = data[overlap:]
      offset = self._next_expected_offset

    self._buffer[offset] = data

  @property
  def next_expected_offset(self) -> int:
    """Returns the next contiguous byte offset expected."""
    return self._next_expected_offset

  def clear(self):
    """Clears the internal cache of out-of-order chunks.

    This must be called when the GRPC stream drops and is retried. It prevents
    stale chunks from a previous dead stream from causing complex overlap
    corruptions or memory leaks if the new stream uses different chunk
    boundaries.
    """
    self._buffer.clear()

  def yield_sequential_chunks(self):
    """Yields all cached chunks that form a contiguous sequence.

    Yields:
      bytes: Contiguous chunks of data.
    """
    while self._next_expected_offset in self._buffer:
      data = self._buffer.pop(self._next_expected_offset)
      self._next_expected_offset += len(data)
      yield data


class AsyncDecoupledWriter:
  """A decoupled thread for writing downloaded chunks to disk.

  A consumer thread that pulls from a bounded queue and commits to disk
  sequentially, avoiding blocking the network ingestion loop.
  """

  def __init__(self, target_stream: io.IOBase):
    self._target_stream = target_stream
    self._queue = queue.Queue(maxsize=2)  # Default max queue size
    self._writer_thread = None

  def start(self):
    pass

  def stop(self):
    pass


def bidi_download_object_optimized(
    gapic_client,
    cloud_resource,
    download_stream,
    digesters,
    progress_callback,
    start_byte,
    end_byte,
    download_strategy,
    decryption_key,
    redirection_handler,
):
  """Optimized Bidi download using Multi-Range Request and Decoupled Writer.

  Currently a stub that falls back to the original bidi_download_object.

  Args:
    gapic_client: The GAPIC API client to interact with GCS using gRPC.
    cloud_resource: See cloud_api.CloudApi.download_object.
    download_stream: Stream to send the object data to.
    digesters: See cloud_api.CloudApi.download_object.
    progress_callback: See cloud_api.CloudApi.download_object.
    start_byte: Starting point for download.
    end_byte: Ending byte number, inclusive, for download.
    download_strategy: Download strategy used to perform the download.
    decryption_key: The decryption key to be used to download the object.
    redirection_handler: The redirection handler for token errors.

  Returns:
    None
  """
  return bidi_download_object(
      gapic_client,
      cloud_resource,
      download_stream,
      digesters,
      progress_callback,
      start_byte,
      end_byte,
      download_strategy,
      decryption_key,
      redirection_handler,
  )



# -*- coding: utf-8 -*- #
# Copyright 2015 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Update cluster command."""

from googlecloudsdk.api_lib.dataproc import constants as dataproc_constants
from googlecloudsdk.api_lib.dataproc import dataproc as dp
from googlecloudsdk.api_lib.dataproc import exceptions
from googlecloudsdk.api_lib.dataproc import util
from googlecloudsdk.calliope import actions
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.dataproc import clusters
from googlecloudsdk.command_lib.dataproc import flags
from googlecloudsdk.command_lib.dataproc.utils import user_sa_mapping_util
from googlecloudsdk.command_lib.util.args import labels_util
from googlecloudsdk.core import log
from googlecloudsdk.core.util import times


@base.UniverseCompatible
class Update(base.UpdateCommand):
  """Update labels and/or the number of worker nodes in a cluster.

  Update the number of worker nodes and/or the labels in a cluster.

  ## EXAMPLES

  To resize a cluster, run:

    $ {command} my-cluster --region=us-central1 --num-workers=5

  To change the number of secondary workers in a cluster, run:

    $ {command} my-cluster --region=us-central1 --num-secondary-workers=5

  To add the label 'customer=acme' to a cluster, run:

    $ {command} my-cluster --region=us-central1 --update-labels=customer=acme

  To update the label 'customer=ackme' to 'customer=acme', run:

    $ {command} my-cluster --region=us-central1 --update-labels=customer=acme

  To remove the label whose key is 'customer', run:

    $ {command} my-cluster --region=us-central1 --remove-labels=customer

  """

  @classmethod
  def Args(cls, parser):
    dataproc = dp.Dataproc(cls.ReleaseTrack())
    base.ASYNC_FLAG.AddToParser(parser)
    # Allow the user to specify new labels as well as update/remove existing
    labels_util.AddUpdateLabelsFlags(parser)
    # Graceful decomissioning timeouts can be up to 24 hours + add 1 hour for
    # deleting VMs, etc.
    flags.AddTimeoutFlag(parser, default='25h')
    flags.AddClusterResourceArg(parser, 'update', dataproc.api_version)
    parser.add_argument(
        '--num-workers',
        type=int,
        help='The new number of worker nodes in the cluster.')
    num_secondary_workers = parser.add_argument_group(mutex=True)
    num_secondary_workers.add_argument(
        '--num-preemptible-workers',
        action=actions.DeprecationAction(
            '--num-preemptible-workers',
            warn=('The `--num-preemptible-workers` flag is deprecated. '
                  'Use the `--num-secondary-workers` flag instead.')),
        type=int,
        hidden=True,
        help='The new number of preemptible worker nodes in the cluster.')
    num_secondary_workers.add_argument(
        '--num-secondary-workers',
        type=int,
        help='The new number of secondary worker nodes in the cluster.')

    parser.add_argument(
        '--graceful-decommission-timeout',
        type=arg_parsers.Duration(lower_bound='0s', upper_bound='1d'),
        help="""
              The graceful decommission timeout for decommissioning Node Managers
              in the cluster, used when removing nodes. Graceful decommissioning
              allows removing nodes from the cluster without interrupting jobs in
              progress. Timeout specifies how long to wait for jobs in progress to
              finish before forcefully removing nodes (and potentially
              interrupting jobs). Timeout defaults to 0 if not set (for forceful
              decommission), and the maximum allowed timeout is 1 day.
              See $ gcloud topic datetimes for information on duration formats.
              """,
    )

    parser.add_argument(
        '--min-secondary-worker-fraction',
        help=(
            'Minimum fraction of new secondary worker nodes added in a scale up'
            ' update operation, required to update the cluster. If it is not'
            ' met, cluster updation will rollback the addition of secondary'
            ' workers. Must be a decimal value between 0 and 1. Defaults to'
            ' 0.0001.'
        ),
        type=float,
    )

    _AddAlphaArguments(parser, cls.ReleaseTrack())

    idle_delete_group = parser.add_mutually_exclusive_group()
    idle_delete_group.add_argument(
        '--max-idle',
        type=arg_parsers.Duration(),
        hidden=True,
        help="""\
        The duration after the last job completes to auto-delete the cluster,
        such as "2h" or "1d".
        See $ gcloud topic datetimes for information on duration formats.
        """)
    idle_delete_group.add_argument(
        '--no-max-idle',
        action='store_true',
        hidden=True,
        help="""\
        Cancels the cluster auto-deletion by cluster idle duration (configured
         by --max-idle flag)
        """,
    )
    idle_delete_group.add_argument(
        '--delete-max-idle',
        type=arg_parsers.Duration(),
        help="""\
        The duration after the last job completes to auto-delete the cluster,
        such as "2h" or "1d".
        See $ gcloud topic datetimes for information on duration formats.
        """)
    idle_delete_group.add_argument(
        '--no-delete-max-idle',
        action='store_true',
        help="""\
        Cancels the cluster auto-deletion by cluster idle duration (configured
        by --delete-max-idle flag)
        """)

    auto_delete_group = parser.add_mutually_exclusive_group()
    auto_delete_group.add_argument(
        '--max-age',
        type=arg_parsers.Duration(),
        hidden=True,
        help="""\
        The lifespan of the cluster, with auto-deletion upon completion,
        "2h" or "1d".
        See $ gcloud topic datetimes for information on duration formats.
        """)
    auto_delete_group.add_argument(
        '--expiration-time',
        type=arg_parsers.Datetime.Parse,
        hidden=True,
        help="""\
        The time when the cluster will be auto-deleted, such as
        "2017-08-29T18:52:51.142Z". See $ gcloud topic datetimes for
        information on time formats.
        """)
    auto_delete_group.add_argument(
        '--no-max-age',
        action='store_true',
        hidden=True,
        help="""\
        Cancels the cluster auto-deletion by maximum cluster age (configured by
         --max-age or --expiration-time flags)
        """)

    auto_delete_group.add_argument(
        '--delete-max-age',
        type=arg_parsers.Duration(),
        help="""\
        The lifespan of the cluster with auto-deletion upon completion,
        such as "2h" or "1d".
        See $ gcloud topic datetimes for information on duration formats.
        """)
    auto_delete_group.add_argument(
        '--delete-expiration-time',
        type=arg_parsers.Datetime.Parse,
        help="""\
        The time when the cluster will be auto-deleted, such as
        "2017-08-29T18:52:51.142Z". See $ gcloud topic datetimes for
        information on time formats.
        """)
    auto_delete_group.add_argument(
        '--no-delete-max-age',
        action='store_true',
        help="""\
        Cancels the cluster auto-deletion by maximum cluster age (configured
        by --delete-max-age or --delete-expiration-time flags)
        """)

    idle_stop_group = parser.add_mutually_exclusive_group()
    idle_stop_group.add_argument(
        '--stop-max-idle',
        type=arg_parsers.Duration(),
        help="""\
        The duration after the last job completes to auto-stop the cluster,
        such as "2h" or "1d".
        See $ gcloud topic datetimes for information on duration formats.
        """)
    idle_stop_group.add_argument(
        '--no-stop-max-idle',
        action='store_true',
        help="""\
        Cancels the cluster auto-stop by cluster idle duration (configured
        by --stop-max-idle flag)
        """)

    auto_stop_group = parser.add_mutually_exclusive_group()
    auto_stop_group.add_argument(
        '--stop-max-age',
        type=arg_parsers.Duration(),
        help="""\
        The lifespan of the cluster, with auto-stop upon completion,
        such as "2h" or "1d".
        See $ gcloud topic datetimes for information on duration formats.
        """)
    auto_stop_group.add_argument(
        '--stop-expiration-time',
        type=arg_parsers.Datetime.Parse,
        help="""\
        The time when the cluster will be auto-stopped, such as
        "2017-08-29T18:52:51.142Z". See $ gcloud topic datetimes for
        information on time formats.
        """)
    auto_stop_group.add_argument(
        '--no-stop-max-age',
        action='store_true',
        help="""\
        Cancels the cluster auto-stop by maximum cluster age (configured by
        --stop-max-age or --stop-expiration-time flags)
        """)

    # Can only specify one of --autoscaling-policy or --disable-autoscaling
    autoscaling_group = parser.add_mutually_exclusive_group()
    flags.AddAutoscalingPolicyResourceArgForCluster(
        autoscaling_group, api_version='v1')
    autoscaling_group.add_argument(
        '--disable-autoscaling',
        action='store_true',
        help="""\
        Disable autoscaling, if it is enabled. This is an alias for passing the
        empty string to --autoscaling-policy'.
        """)
    user_sa_mapping_util.AddUpdateUserSaMappingFlags(parser)

  def Run(self, args):
    dataproc = dp.Dataproc(self.ReleaseTrack())
    cluster_ref = args.CONCEPTS.cluster.Parse()

    cluster_config = dataproc.messages.ClusterConfig()
    changed_fields = []

    has_changes = False
    no_update_error_msg = (
        'Must specify at least one cluster parameter to update.'
    )

    if args.num_workers is not None:
      worker_config = dataproc.messages.InstanceGroupConfig(
          numInstances=args.num_workers)
      cluster_config.workerConfig = worker_config
      changed_fields.append('config.worker_config.num_instances')
      has_changes = True

    num_secondary_workers = _FirstNonNone(args.num_preemptible_workers,
                                          args.num_secondary_workers)
    if num_secondary_workers is not None:
      worker_config = dataproc.messages.InstanceGroupConfig(
          numInstances=num_secondary_workers)
      cluster_config.secondaryWorkerConfig = worker_config
      changed_fields.append(
          'config.secondary_worker_config.num_instances')
      has_changes = True

    if args.min_secondary_worker_fraction is not None:
      if cluster_config.secondaryWorkerConfig is None:
        worker_config = dataproc.messages.InstanceGroupConfig(
            startupConfig=dataproc.messages.StartupConfig(
                requiredRegistrationFraction=(
                    args.min_secondary_worker_fraction
                )
            )
        )
      else:
        worker_config = dataproc.messages.InstanceGroupConfig(
            numInstances=num_secondary_workers,
            startupConfig=dataproc.messages.StartupConfig(
                requiredRegistrationFraction=(
                    args.min_secondary_worker_fraction
                )
            ),
        )
      cluster_config.secondaryWorkerConfig = worker_config
      changed_fields.append(
          'config.secondary_worker_config.startup_config.required_registration_fraction'
      )
      has_changes = True

    if self.ReleaseTrack() == base.ReleaseTrack.ALPHA:
      if args.secondary_worker_standard_capacity_base is not None:
        if cluster_config.secondaryWorkerConfig is None:
          worker_config = dataproc.messages.InstanceGroupConfig(
              instanceFlexibilityPolicy=dataproc.messages.InstanceFlexibilityPolicy(
                  provisioningModelMix=dataproc.messages.ProvisioningModelMix(
                      standardCapacityBase=args.secondary_worker_standard_capacity_base
                  )))
        else:
          worker_config = dataproc.messages.InstanceGroupConfig(
              numInstances=num_secondary_workers,
              startupConfig=cluster_config.secondaryWorkerConfig.startupConfig,
              instanceFlexibilityPolicy=dataproc.messages.InstanceFlexibilityPolicy(
                  provisioningModelMix=dataproc.messages.ProvisioningModelMix(
                      standardCapacityBase=args.secondary_worker_standard_capacity_base
                  )
              )
          )
        cluster_config.secondaryWorkerConfig = worker_config
        changed_fields.append(
            'config.secondary_worker_config.instance_flexibility_policy.provisioning_model_mix.standard_capacity_base'
        )
        has_changes = True

    if args.autoscaling_policy:
      cluster_config.autoscalingConfig = dataproc.messages.AutoscalingConfig(
          policyUri=args.CONCEPTS.autoscaling_policy.Parse().RelativeName())
      changed_fields.append('config.autoscaling_config.policy_uri')
      has_changes = True
    elif args.autoscaling_policy == '' or args.disable_autoscaling:  # pylint: disable=g-explicit-bool-comparison
      # Disabling autoscaling. Don't need to explicitly set
      # cluster_config.autoscaling_config to None.
      changed_fields.append('config.autoscaling_config.policy_uri')
      has_changes = True

    lifecycle_config = dataproc.messages.LifecycleConfig()
    changed_config = False
    # Flags max_age, expiration_time, max_idle, no_max_age, no_max_idle are
    # hidden, but still supported. They are replaced with new flags
    # delete_max_age, delete_expiration_time, delete_max_idle,
    # no_delete_max_age and no_delete_max_idle.
    if args.max_age is not None:
      lifecycle_config.autoDeleteTtl = str(args.max_age) + 's'
      changed_fields.append('config.lifecycle_config.auto_delete_ttl')
      changed_config = True
    if args.expiration_time is not None:
      lifecycle_config.autoDeleteTime = times.FormatDateTime(
          args.expiration_time)
      changed_fields.append('config.lifecycle_config.auto_delete_time')
      changed_config = True
    if args.max_idle is not None:
      lifecycle_config.idleDeleteTtl = str(args.max_idle) + 's'
      changed_fields.append('config.lifecycle_config.idle_delete_ttl')
      changed_config = True
    if args.no_max_age:
      lifecycle_config.autoDeleteTtl = None
      changed_fields.append('config.lifecycle_config.auto_delete_ttl')
      changed_config = True
    if args.no_max_idle:
      lifecycle_config.idleDeleteTtl = None
      changed_fields.append('config.lifecycle_config.idle_delete_ttl')
      changed_config = True

    if args.delete_max_age is not None:
      lifecycle_config.autoDeleteTtl = str(args.delete_max_age) + 's'
      changed_fields.append('config.lifecycle_config.auto_delete_ttl')
      changed_config = True
    if args.delete_expiration_time is not None:
      lifecycle_config.autoDeleteTime = times.FormatDateTime(
          args.delete_expiration_time
      )
      changed_fields.append('config.lifecycle_config.auto_delete_time')
      changed_config = True
    if args.delete_max_idle is not None:
      lifecycle_config.idleDeleteTtl = str(args.delete_max_idle) + 's'
      changed_fields.append('config.lifecycle_config.idle_delete_ttl')
      changed_config = True
    if args.no_delete_max_age:
      lifecycle_config.autoDeleteTtl = None
      changed_fields.append('config.lifecycle_config.auto_delete_ttl')
      changed_config = True
    if args.no_delete_max_idle:
      lifecycle_config.idleDeleteTtl = None
      changed_fields.append('config.lifecycle_config.idle_delete_ttl')
      changed_config = True

    if args.stop_max_age is not None:
      lifecycle_config.autoStopTtl = str(args.stop_max_age) + 's'
      changed_fields.append('config.lifecycle_config.auto_stop_ttl')
      changed_config = True
    if args.stop_expiration_time is not None:
      lifecycle_config.autoStopTime = times.FormatDateTime(
          args.stop_expiration_time)
      changed_fields.append('config.lifecycle_config.auto_stop_time')
      changed_config = True
    if args.stop_max_idle is not None:
      lifecycle_config.idleStopTtl = str(args.stop_max_idle) + 's'
      changed_fields.append('config.lifecycle_config.idle_stop_ttl')
      changed_config = True
    if args.no_stop_max_age:
      lifecycle_config.autoStopTtl = None
      changed_fields.append('config.lifecycle_config.auto_stop_ttl')
      changed_config = True
    if args.no_stop_max_idle:
      lifecycle_config.idleStopTtl = None
      changed_fields.append('config.lifecycle_config.idle_stop_ttl')
      changed_config = True

    if changed_config:
      cluster_config.lifecycleConfig = lifecycle_config
      has_changes = True

    # Put in a thunk so we only make this call if needed
    def _GetCurrentLabels():
      # We need to fetch cluster first so we know what the labels look like. The
      # labels_util will fill out the proto for us with all the updates and
      # removals, but first we need to provide the current state of the labels
      current_cluster = _GetCurrentCluster(dataproc, cluster_ref)
      return current_cluster.labels
    labels_update = labels_util.ProcessUpdateArgsLazy(
        args, dataproc.messages.Cluster.LabelsValue,
        orig_labels_thunk=_GetCurrentLabels)
    if labels_update.needs_update:
      has_changes = True
      changed_fields.append('labels')
    labels = labels_update.GetOrNone()

    def _GetCurrentUserServiceAccountMapping():
      current_cluster = _GetCurrentCluster(dataproc, cluster_ref)
      if (
          current_cluster.config.securityConfig
          and current_cluster.config.securityConfig.identityConfig
      ):
        return (
            current_cluster.config.securityConfig.identityConfig.userServiceAccountMapping
        )
      return None

    def _UpdateSecurityConfig(cluster_config, user_sa_mapping):
      if cluster_config.securityConfig is None:
        cluster_config.securityConfig = dataproc.messages.SecurityConfig()
      if cluster_config.securityConfig.identityConfig is None:
        cluster_config.securityConfig.identityConfig = (
            dataproc.messages.IdentityConfig()
        )

      cluster_config.securityConfig.identityConfig.userServiceAccountMapping = (
          user_sa_mapping
      )

    if args.add_user_mappings or args.remove_user_mappings:
      if not _IsMultitenancyCluster(_GetCurrentCluster(dataproc, cluster_ref)):
        raise exceptions.ArgumentError(
            'User service account mapping can only be updated for multi-tenant'
            ' clusters.'
        )
      user_sa_mapping_update = user_sa_mapping_util.ProcessUpdateArgsLazy(
          args,
          dataproc.messages.IdentityConfig.UserServiceAccountMappingValue,
          orig_user_sa_mapping_thunk=_GetCurrentUserServiceAccountMapping,
      )
      if user_sa_mapping_update.needs_update:
        changed_fields.append(
            'config.security_config.identity_config.user_service_account_mapping'
        )
        has_changes = True
      else:
        if args.add_user_mappings:
          no_update_error_msg += (
              ' User to add is already present in service account mapping.'
          )
        if args.remove_user_mappings:
          no_update_error_msg += (
              ' User to remove is not present in service account mapping.'
          )
      user_sa_mapping = user_sa_mapping_update.GetOrNone()
      if user_sa_mapping:
        _UpdateSecurityConfig(cluster_config, user_sa_mapping)
    elif args.identity_config_file:
      if not _IsMultitenancyCluster(_GetCurrentCluster(dataproc, cluster_ref)):
        raise exceptions.ArgumentError(
            'User service account mapping can only be updated for multi-tenant'
            ' clusters.'
        )
      if cluster_config.securityConfig is None:
        cluster_config.securityConfig = dataproc.messages.SecurityConfig()
      cluster_config.securityConfig.identityConfig = (
          clusters.ParseIdentityConfigFile(dataproc, args.identity_config_file)
      )
      changed_fields.append(
          'config.security_config.identity_config.user_service_account_mapping'
      )
      has_changes = True

    if not has_changes:
      raise exceptions.ArgumentError(no_update_error_msg)

    cluster = dataproc.messages.Cluster(
        config=cluster_config,
        clusterName=cluster_ref.clusterName,
        labels=labels,
        projectId=cluster_ref.projectId)

    request = dataproc.messages.DataprocProjectsRegionsClustersPatchRequest(
        clusterName=cluster_ref.clusterName,
        region=cluster_ref.region,
        projectId=cluster_ref.projectId,
        cluster=cluster,
        updateMask=','.join(changed_fields),
        requestId=util.GetUniqueId())

    if args.graceful_decommission_timeout is not None:
      request.gracefulDecommissionTimeout = (
          str(args.graceful_decommission_timeout) + 's'
      )

    operation = dataproc.client.projects_regions_clusters.Patch(request)

    if args.async_:
      log.status.write(
          'Updating [{0}] with operation [{1}].'.format(
              cluster_ref, operation.name))
      return

    util.WaitForOperation(
        dataproc,
        operation,
        message='Waiting for cluster update operation',
        timeout_s=args.timeout)

    request = dataproc.messages.DataprocProjectsRegionsClustersGetRequest(
        projectId=cluster_ref.projectId,
        region=cluster_ref.region,
        clusterName=cluster_ref.clusterName)
    cluster = dataproc.client.projects_regions_clusters.Get(request)
    log.UpdatedResource(cluster_ref)
    return cluster


def _FirstNonNone(first, second):
  return first if first is not None else second


def _AddAlphaArguments(parser, release_track):

  if release_track == base.ReleaseTrack.ALPHA:

    parser.add_argument(
        '--secondary-worker-standard-capacity-base',
        type=int,
        help="""
              The number of standard VMs in the Spot and Standard Mix
        feature.
              """,
    )


def _IsMultitenancyCluster(cluster) -> bool:
  """Checks if the cluster is a multi-tenant cluster.

  Args:
    cluster: The cluster configuration.

  Returns:
    True if the cluster is a multi-tenant cluster, False otherwise.
  """
  config = cluster.config
  if config and config.softwareConfig and config.softwareConfig.properties:
    props = config.softwareConfig.properties
    for prop in props.additionalProperties:
      if (
          prop.key == dataproc_constants.ENABLE_DYNAMIC_MULTI_TENANCY_PROPERTY
          and prop.value.lower() == 'true'
      ):
        return True
  return False


def _GetCurrentCluster(dataproc, cluster_ref):
  """Retrieves the current cluster configuration.

  Args:
    dataproc: The Dataproc API client.
    cluster_ref: The reference to the cluster.

  Returns:
    The current cluster configuration.
  """
  # This is used for labels and auxiliary_node_pool_configs
  get_cluster_request = (
      dataproc.messages.DataprocProjectsRegionsClustersGetRequest(
          projectId=cluster_ref.projectId,
          region=cluster_ref.region,
          clusterName=cluster_ref.clusterName,
      )
  )
  current_cluster = dataproc.client.projects_regions_clusters.Get(
      get_cluster_request
  )
  return current_cluster
